    
    def analyze_pgv_chi2(self, field, fnl, bv, nkbins, ddof):
        """Do model parameters (fnl,bv) fit the data? Returns (chi2, ndof, p-value).

        The 'field' argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
           - field=[1,0] for 90 GHz
           - field=[0,1] for 150 GHz
           - field=[1,-1] for null (90-150) GHz.

        The 'nkbins' argument is the number of k-bins used to compute the chi^2.

        The 'ddof' argument is used to compute the number of degrees of freedom:
            ndof = nkbins - ddof
        """

        d = self.pgv_data(field)[:nkbins]              # shape (nkbins,)
        s = self._pgv_surr(field, bv, fnl)[:,:nkbins]  # shape (nsurr,nkbins)
        x = d - np.mean(s, axis=0)                     # shape (nkbins,)
        cov = np.cov(s, rowvar=False)                  # shape (nkbins,nkbins)
        
        chi2 = np.dot(x, np.linalg.solve(cov,x))
        ndof = nkbins - ddof
        pte = scipy.stats.chi2.sf(chi2, ndof)
        
        return chi2, ndof, pte




from .. import Box
from . import helpers

import numpy as np

    
def test_k_component():
    print(f'test_k_component(): start')
    
    for iouter in range(100):
        box = helpers.random_box()
        axis = np.random.randint(0, box.ndim)
        zero_nyquist = (np.random.uniform() < 0.5)

        n = box.npix[axis]
        k = box.get_k_component(axis, zero_nyquist=zero_nyquist, one_dimensional=True)
        
        k2 = np.array([ i if (2*i <= n) else (i-n) for i in range(box.nk[axis]) ])
        k2 = k2 * (2*np.pi / box.boxsize[axis])
        
        if zero_nyquist and (n % 2 == 0):
            k2[n//2] = 0.

        eps = box.boxsize[axis] * np.max(np.abs(k-k2))
        assert eps < 1.0e-10
        
    print(f'test_k_component(): pass')


def test_r_component():
    print(f'test_r_component(): start')
    
    for iouter in range(100):
        box = helpers.random_box()
        axis = np.random.randint(0, box.ndim)
        r = box.get_r_component(axis, one_dimensional=True)
        r2 = np.linspace(box.lpos[axis], box.rpos[axis], box.npix[axis])
        eps = np.max(np.abs(r-r2))
        assert eps < 1.0e-10
        
    print(f'test_r_component(): pass')


def test_smallest_r():
    print(f'test_smallest_r(): start')

    for iouter in range(100):
        box = helpers.random_box()
        r = box.get_r()
        rmin = np.min(r)
        rmin2 = r[box._ix_smallest_r]
        eps = np.abs(rmin-rmin2)
        assert eps < 1.0e-10
        
    print(f'test_smallest_r(): pass')




"""This source file is intended to be a home for plotting-related utility functions, but there's not much here yet!"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib import cm


def savefig(filename=None, tight_layout=True):
    if tight_layout:
        plt.tight_layout()
    
    if filename is None:
        plt.show()
    else:
        print(f'Writing {filename}\n', end='')
        plt.savefig(filename)
    
    plt.clf()

    
def real_space_plotter(arr, filename=None, axis=None, vmax=None, title=None):
    r"""General purpose function for plotting a real-space field by averaging over specified axes.
    
    Function args:
       - ``arr`` (array-like): the real-space field to plot
       - ``filename`` (str, optional): filename for saving the plot as a PDF
       - ``axis`` (list of ints, optional): the axes to average over for each plot
       - ``vmax`` (float, optional): maximum color scale value; if None, determined automatically
       - ``title``: (str, optional): title for the plot

    This function displays and optionally saves the plot with symmetric color scaling around zero (from -vmax to vmax).

    (Source: Selim's ``helperfunctions.py``, Nov 2024.)
    """

    arr = np.asarray(arr)
    
    if arr.dtype != float:
        raise RuntimeError(f'kszx.plot.real_space_plotter: expected floating-point array, got {arr.dtype=}')
    
    if axis is None:
        axis = range(arr.ndim)

    axis = [ int(x) for x in axis ]  # convert to list of ints
    assert all(0 <= d < arr.ndim for d in axis)
        
    # Initialize figure and axes for subplots, with a number of columns equal to the number of axes specified
    fig, axs = plt.subplots(1, len(axis), figsize=(4 * len(axis), 4), sharey=True, sharex=True)
    
    # Compute the mean of the real part of the matrix along each specified axis for plotting
    to_plot = [arr.mean(axis=ax) for ax in axis]
    
    # Set vmax for color scale if not provided
    if vmax is None:
        vmax = max(np.max(np.abs(x)) for x in to_plot)

    if title is not None:
        plt.title(title)
    
    # Generate images on each subplot for the averaged data along each specified axis
    im = [axs[ax].imshow(to_plot[ax], cmap=cm.RdBu_r, vmax=vmax, vmin=-vmax) for ax in range(len(axis))]

    # Add color bar to the right of the last axis (assumed to be 2)
    divider = make_axes_locatable(axs[-1])
    cax = divider.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im[-1], cax=cax, orientation='vertical')
        
    # Adjust layout to avoid overlaps
    plt.tight_layout()

    # Save or show plot (depending on whether filename is None)
    if filename is not None:
        print(f'Writing {filename}\n', end='')
        plt.savefig(filename)
    else:
        plt.show()



import os
import sys
import h5py
import gzip
import uuid
import shutil
import fitsio
import pickle   # FIXME cPickle?
import tarfile
import zipfile
import functools
import subprocess
import numpy as np

# Avoid name collision with io_utils.wget() below.
import wget as wget_module


def read_pickle(filename):
    print(f'Reading {filename}\n', end='')
    with open(filename, 'rb') as f:
        return pickle.load(f)


def write_pickle(filename, x):
    mkdir_containing(filename)
    print(f'Writing {filename}\n', end='') 
    with open(filename, 'wb') as f:
        pickle.dump(x, f)


def read_npy(filename):
    print(f'Reading {filename}\n', end='')
    return np.load(filename, allow_pickle=True)


def write_npy(filename, arr):
    mkdir_containing(filename)
    print(f'Writing {filename}\n', end='') 
    np.save(filename, arr)

    
def read_npz(filename):
    """Returns a dictionary-like object (NpzFile) which maps names to arrays.

    Some arrays may have special names 'arr_0', 'arr_1', ... 
    This happens for arrays which were positional (not keyword) arguments to write_npz().
    """

    # read_npy() and read_npz() can be the same function,
    # since np.load() works for both .npy and .npz files.
    return read_npy(filename)


def write_npz(filename, *args, **kwds):
    """Example: write_npz('file.npz', a, b, xarr=x, yarr=y), where (a,b,x,y) are numpy arrays.

    In this example, when 'file.npz' is read (with np.load() or kszx.read_npz()), the return 
    value is a dictionary-like object (NpzFile) with keys 'arr_0', 'arr_1', 'xarr', 'yarr'.
    Note that the positional arguments (a,b) get special names 'arr_0', 'arr_1'.
    """

    mkdir_containing(filename)
    print(f'Writing {filename}\n', end='')
    np.savez(filename, *args, **kwds)


####################################################################################################


@functools.cache
def get_data_dir():
    """Returns root directory for auto-downloaded kszx data. (Either $KSZX_DATA_DIR or $HOME/kszx_data.)"""

    if 'KSZX_DATA_DIR' in os.environ:
        data_dir = os.environ['KSZX_DATA_DIR']
        print(f'kszx: using environment variable $KSZX_DATA_DIR: {data_dir}')
    elif 'HOME' in os.environ:
        data_dir = os.path.join(os.environ['HOME'], 'kszx_data')
        print(f'kszx: environment variable $KSZX_DATA_DIR not defined, using {data_dir} instead')
    else:
        raise RuntimeError('kszx.get_data_dir(): neither $KSZX_DATA_DIR nor $HOME environment variables were defined?!')

    if os.path.isdir(data_dir):
        return data_dir
    
    raise RuntimeError(f'kszx: data directory {data_dir} not found.\n'
                       + f"This is the directory where kszx will auto-download and retrieve data files.\n"
                       + f"Here are some options:\n"
                       + f"  1. if you want to store data in this directory, do 'mkdir -p {data_dir}'\n"
                       + f"  2. if you want to use a different directory, do 'ln -s /some/other/dir {data_dir}\n"
                       + f"  3. if you want to use a different directory, another option is to set the $KSZX_DATA_DIR env variable\n"
                       + f"For more discussion, see: https://kszx.readthedocs.io/en/latest/intro.html#data-files")


def mkdir_containing(filename):
    assert isinstance(filename, str)
    
    d = os.path.dirname(filename)
    if (d == '') or os.path.exists(d):
        return
    
    print(f'Creating directory {d}')
    os.makedirs(d, exist_ok=True)
    

def do_download(abspath, download=False, dlfunc=None):
    """This short helper function appears in many functions which decide whether to download a file.

    Here and in other parts of kszx, the 'dlfunc' argument gives the name of a transitive caller that
    expects the file to be present, and has a 'download=False' optional argument. This information is
    only used when generating exception-text (to tell the user how to download the file).
    """
    
    if os.path.exists(abspath):
        return False
    if download:
        return True
    if dlfunc is not None:
        raise RuntimeError(f'File {abspath} not found. To auto-download, call {dlfunc}() with download=True.')
    return False


def wget(filename, url):
    if os.path.exists(filename):
        print(f'File {filename} already exists -- skipping download')
        return
    
    mkdir_containing(filename)
    print(f'Downloading {url} -> {filename}')
    
    wget_module.download(url, out=filename)
    print()   # extra newline after progress bar
    assert os.path.exists(filename)


def unpack(srcfile, expected_dstfile=None):
    """Unpacks a file ending in one of: .gz, .tar, .tgz, .tar.gz"""
    
    if not os.path.exists(srcfile):
        raise RuntimeError(f"kszx.io_utils.unpack(): source file '{srcfile}' does not exist")
    
    if srcfile.endswith('.tar'):
        print(f'Un-tarring {srcfile}')
        tar = tarfile.open(srcfile, 'r:')
        tar.extractall(path = os.path.dirname(srcfile))
        tar.close()
    
    elif srcfile.endswith('.tar.gz') or srcfile.endswith('.tgz'):
        print(f'Un-tarring and gunzipping {srcfile}')
        tar = tarfile.open(srcfile, 'r:gz')
        tar.extractall(path = os.path.dirname(srcfile))
        tar.close()
    
    elif srcfile.endswith('.gz'):
        dstfile = srcfile[:-3]
        print(f'Gunzipping {srcfile} -> {dstfile}')
        with gzip.open(srcfile, 'rb') as f_in:
            with open(dstfile, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

    elif srcfile.endswith('.zip'):
        print(f'Unzipping {srcfile}')
        with zipfile.ZipFile(srcfile, 'r') as f:
            f.extractall(os.path.dirname(srcfile))

    else:
        raise RuntimeError(f"kszx.io_utils.unpack(): source file '{srcfile}' must end in .zip, .gz, .tgz, or .tar.gz")

    if (expected_dstfile is not None) and not os.path.exists(expected_dstfile):
        raise RuntimeError(f"kszx.io_utils.unpack(): after unpacking {srcfile=}, {expected_dstfile=} does not exist")

    
def nersc_scp_download(remote_abspath, local_abspath):
    """A temporary kludge, since ACT DR6 maps are currently on NERSC but not LAMBDA.

    Documentation by example:

      nersc_scp_download(
        remote_abspath = '/global/cfs/cdirs/cmb/data/act_dr6/dr6.02/maps/published/act-planck_dr6.02_coadd_AA_daynight_f090_map_srcfree.fits',
        local_abspath = '/home/kmsmith/kszx_data/act/dr6/maps/published/act-planck_dr6.02_coadd_AA_daynight_f090_map_srcfree.fits'
      )
    """
    
    if os.path.exists(local_abspath):
        return

    print(f"File {local_abspath} is not currently available on the web -- the only option is a NERSC scp download.")

    mkdir_containing(local_abspath)
    tmpfile = local_abspath +  f'.tmp{uuid.uuid4().hex[:8]}'
    ssh_key = os.path.join(os.environ['HOME'], '.ssh', 'nersc')
    ssh_args = ['scp', '-i', ssh_key, f'perlmutter.nersc.gov:{remote_abspath}', tmpfile ]
    
    if not os.path.exists(ssh_key):
        raise RuntimeError('To download files from NERSC, you need to have a NERSC account, and run sshproxy.sh. '
                           + 'See https://docs.nersc.gov/connect/mfa/#sshproxy')
    
    try:
        print(f"Executing command: {' '.join(ssh_args)}")
        subprocess.run(ssh_args, check=True)
    
    except subprocess.CalledProcessError as e:
        print(f"Couldn't download from NERSC. One possible reason is that your ssh key is out of date, "
              + f"and you need to rerun sshproxy.sh. (See https://docs.nersc.gov/connect/mfa/#sshproxy). "
              + f"Error message was: {e}")

    if not os.path.exists(tmpfile):
        raise RuntimeError(f'scp command succeeded, but destination file {tmpfile} was not created?!')

    print(f'Renaming {tmpfile} -> {local_abspath}')
    os.rename(tmpfile, local_abspath)


####################################################################################################


def show_file(filename, file_type=None, file=None):
    """Prints (to stdout) a simple human-readable summary of file with given name.

    The file_type can be one of:

      pkl        python pickle
      npy        numpy array
      npz        numpy NpzFile
      h5, hdf5   HDF5 file
      fits       FITS file

    If file_type is None, then the file_type is determined from the filename in 
    a brain-dead way (e.g. if filename='x.pkl', then file_type='pkl').

    Can be invoked from command line with 'python -m kszx show <filename>'.
    """

    assert isinstance(filename, str)
        
    if file_type is None:
        b = os.path.basename(filename)
        i = b.rfind('.')
        if (i < 0) or (i >= len(b)-1):
            raise RuntimeError(f"Couldn't determine file_type from {filename=}")
        file_type = b[(i+1):]

    if file_type == 'pkl':
        x = read_pickle(filename)
        print(x, file=file)

    elif file_type == 'npy':
        x = read_npy(filename)
        print('Shape:', x.shape)
        print(x, file=file)

    elif file_type == 'npz':
        _show_npz(filename, file=file)

    elif file_type in [ 'h5', 'hdf5' ]:
        _show_h5(filename, file=file)

    elif file_type in [ 'fits' ]:
        _show_fits(filename, file=file)
            
    else:
        raise RuntimeError(f"kszx.show_file(): unsupported file_type {file_type}")
        

def _show_npz(x, file=None, indent='  '):
    """Intended to be called through show_file(), or via command line.

    Prints (to stdout) a simple human-readable summary of 'x', which can be
    either a filename or a NpzFile (class numpy.lib.npyio.NpzFile).
    """

    if isinstance(x, str):
        x = read_npz(x)

    if not isinstance(x, np.lib.npyio.NpzFile):
        raise RuntimeError('kszx._show_npz(): argument must be either string or NpzFile')

    print(x, file=file)
    
    for k,v in x.items():
        print(f'{indent}{k}: shape={v.shape}, min={np.min(v)}, max={np.max(v)}', file=file)


def _show_h5_attrs(x, prefix, file):
    for k,v in sorted(x.attrs.items()):
        print(f"{prefix}Attribute '{k}': {v}", file=file)
    

def _show_h5(x, prefix=None, file=None, indent='  '):
    """Intended to be called through show_file(), or via command line.

    Prints (to stdout) a simple human-readable summary of 'x', which can be
    either a filename, an h5py.File, or an h5py.Group.

    The default value of the 'indent' argument indents by two spaces at every level.
    """

    if prefix is None:
        prefix = ''
    if file is None:
        file = sys.stdout
    
    if isinstance(x, str):
        with h5py.File(x) as f:
            print(f"{prefix}File '{x}'", file=file)
            _show_h5(f, prefix + indent, file, indent)
        return

    _show_h5_attrs(x, prefix, file)

    # x.items() = "all groups and datasets in a single dict".
    # First pass: loop over datasets.
    
    for k,v in sorted(x.items()):
        if isinstance(v, h5py.Dataset):
            try:
                s = f', min={np.min(v)}, max={np.max(v)}'
            except:
                s = ''
            print(f"{prefix}Dataset '{k}': shape={v.shape}, dtype={v.dtype}{s}", file=file)
            _show_h5_attrs(v, prefix + indent, file)
        elif not isinstance(v, h5py.Group):
            print(f"{prefix}Unrecognized element '{k}': type={type(v)}", file=file)

    # Second pass: loop over groups

    for k,v in sorted(x.items()):
        if isinstance(v, h5py.Group):
            print(f"{prefix}Group '{k}'", file=file)
            _show_h5(v, prefix + indent, file, indent)


def _show_fits(x, file=None):
    """Intended to be called through show_file(), or via command line.

    Prints (to stdout) a simple human-readable summary of 'x', which can be
    either a filename, or a fitsio.FITS object.

    This quick-and-dirty implementation just calls __str__() functions in the
    fitsio python module (which is pretty good).
    """

    if isinstance(x, str):
        with fitsio.FITS(x) as f:
            _show_fits(f)

    elif isinstance(x, fitsio.FITS):
        for hdu in x:
            print(hdu, file=file)

    else:
        raise RuntimeError('kszx._show_fits(): argument must be a string, or of type fitsio.FITS')



import numpy as np
import pixell.enmap
import pixell.enplot
import pixell.curvedsky

from . import utils
from . import io_utils

from .Catalog import Catalog


def read_map(filename):
    print(f'Reading {filename}\n', end='')
    assert filename.endswith('.fits')
    return pixell.enmap.read_map(filename)


def write_map(filename, m):
    # I learned the hard way that you need this assert! If pixell.enmap.write_map() is
    # called with a numpy array (instead of a pixell ndmap), it will write a "time-bomb"
    # WCS which causes some pixell functions to fail later, with cryptic error messages.
    assert isinstance(m, pixell.enmap.ndmap)
    
    print(f'Writing {filename}\n', end='')
    assert filename.endswith('.fits')
    io_utils.mkdir_containing(filename)
    pixell.enmap.write_map(filename, m)


def plot_map(m, downgrade, nolabels=True, filename=None, **kwds):
    """Thin wrapper around pixell.enplot(), just adding a few tweaks:

       - Make downgrade argument mandatory (usually needed in practice to avoid
         runaway heavyweight behavior). Suggest downgrade=40 for ACT.

       - Make nolabels=True the default (haven't figured out how to make labels
         look reasonable).

       - Add 'filename' argument to select between show/write.

    For more kwds, see:
      https://pixell.readthedocs.io/en/latest/reference.html#pixell.enplot.plot"""
    
    assert isinstance(m, pixell.enmap.ndmap)

    bunch = pixell.enplot.plot(m, downgrade=downgrade, nolabels=nolabels, **kwds)

    if filename is None:
        pixell.enplot.show(bunch)
    else:
        print(f'Writing {filename}\n', end='')
        assert filename.endswith('.png')
        io_utils.mkdir_containing(filename)
        pixell.enplot.write(filename[:-4], bunch)


####################################################################################################


def _accum_catalog(ret, shape, wcs, catalog, weights, allow_outliers, normalize_sum=None):
    """Helper for map_from_catalog(). Returns sum of weights."""

    if not allow_outliers:
        # Note: if allow_outliers is False, then pixell_utils.ang2pix() returns (idec, ira).
        msg = "kszx.pixell_utils.map_from_catalog(): some points were out-of-bounds, you may want allow_outliers=True"
        idec, ira = ang2pix(shape, wcs, catalog.ra_deg, catalog.dec_deg, allow_outliers=False, outlier_errmsg=msg)
    else:
        # Note: if allow_outliers is True, then pixell_utils.ang2pix() returns (idec, ira, mask).
        idec, ira, mask = ang2pix(shape, wcs, catalog.ra_deg, catalog.dec_deg, allow_outliers=True)
        idec, ira = idec[mask], ira[mask]
        wvec_flag = (weights is not None) and (weights.ndim > 0)
        weights = weights[mask] if wvec_flag else weights

    ret_1d = np.reshape(ret, (-1,))
    ix_1d = idec*shape[1] + ira

    # Return sum of weights, after applying outlier mask, but before applying 'normalize_sum'.
    return utils.scattered_add(ret_1d, ix_1d, weights, normalize_sum=normalize_sum)
    

def map_from_catalog(shape, wcs, gcat, weights=None, rcat=None, rweights=None, normalized=True, allow_outliers=True):
    """Returns a pixell.enmap.ndmap.
        
    The (rcat, rweights) arguments represent "randoms" to be subtracted.
    NOTE: the rweights are renormalized so that sum(rweights) = -sum(weights)!

    If 'allow_outliers' is True, then "out-of-bounds" galaxies (i.e. galaxies outside 
    the boundaries of the pixell map) will be silently discarded. If 'allow_outliers'
    is False, then out-of-bounds galaxies will raise an exception.

    If normalized=True, then the output map includes a factor 1 / (pixel area).
    This normalization best represents a sum of delta functions f(x) = sum_j w_j delta^2(th_j).
    """

    assert isinstance(gcat, Catalog)
    assert isinstance(rcat, Catalog) or (rcat is None)

    if (rcat is None) and (rweights is not None):
        raise RuntimeError("kszx.healpix_utils.map_from_catalog(): 'rcat' arg is None, but 'rweights' arg is not None")

    weights = utils.asarray(weights, 'kszx.healpix_utils.map_from_catalog', 'weights', dtype=float, allow_none=True)
    rweights = utils.asarray(rweights, 'kszx.healpix_utils.map_from_catalog', 'rweights', dtype=float, allow_none=True)

    assert (weights is None) or (weights.ndim == 0) or (weights.shape == (gcat.size,))
    assert (rweights is None) or (rweights.ndim == 0) or (rweights.shape == (rcat.size,))

    ret = np.zeros(shape)
    wsum = _accum_catalog(ret, shape, wcs, gcat, weights, allow_outliers)

    if rcat is not None:
        _accum_catalog(ret, shape, wcs, rcat, rweights, allow_outliers, normalize_sum = -wsum)

    if normalized:
        ret /= pixell.enmap.pixsizemap(shape, wcs)
    
    return pixell.enmap.enmap(ret, wcs=wcs)


def eval_map_on_catalog(m, catalog, pad=None, return_mask=False):
    """Evaluates pixell map 'm' at (ra,dec) values from catalog, and returns a 1-d array.

    Since pixell maps can be partial-sky, some of the (ra,dec) pairs may be "outliers",
    i.e. outside the map footprint. If pad=None (the default), this raises an exception.
    If 'pad' is specified (e.g. pad=0.0) then outlier values are replaced by 'pad'.

    If return_mask=False (the default), returns a 1-d array 'map_vals'.

    If return_mask=True, returns (map_vals, mask), where 'mask' is a 1-d boolean array which
    is True for non-outliers, False for outliers.
    """

    assert isinstance(m, pixell.enmap.ndmap)
    assert isinstance(catalog, Catalog)

    if return_mask and (pad is None):
        raise RuntimeError("kszx.pixell_utils.eval_map_on_catalog(): if return_mask=True, then 'pad' must be specified")
    if pad is not None:
        pad = float(pad)
    
    # pixell_utils.ang2pix() is defined later in this file.
    # It returns (idec,ira) if allow_outliers is False, or (idec,ira,mask) if allow_outliers is True.

    msg = "kszx.pixell_utils.eval_map_on_catalog(): some points are out-of-bounds, you may want to specify the 'pad' arugment"
    t = ang2pix(m.shape, m.wcs, catalog.ra_deg, catalog.dec_deg, allow_outliers=(pad is not None), outlier_errmsg=msg)
    
    idec, ira = t[0], t[1]
    mask = t[2] if (len(t) > 2) else None
    map_vals = m[idec,ira]

    if pad is not None:
        map_vals = np.where(mask, map_vals, pad)

    return (map_vals, mask) if return_mask else map_vals
    

####################################################################################################


def alm2map(alm, shape, wcs):
    # FIXME map2alm() will blow up with a cryptic error message if it gets a bad WCS (e.g. zeroed).
    # Should find a general test for WCS badness, and add asserts throughout this file.
    
    alm = np.asarray(alm)
    assert alm.ndim == 1
    assert (alm.dtype == complex) or (alm.dtype == np.complex64)

    m = pixell.enmap.enmap(np.zeros(shape), wcs=wcs)
    pixell.curvedsky.alm2map(alm, m)
    return m
    

def map2alm(m, lmax):
    assert isinstance(m, pixell.enmap.ndmap)
    assert lmax >= 0
    return pixell.curvedsky.map2alm(m, lmax=lmax)


####################################################################################################


def ang2pix(shape, wcs, ra_deg, dec_deg, allow_outliers=False, outlier_errmsg=None):
    """Returns (idec,ira) if allow_outliers is False, or (idec,ira,mask) if allow_outliers is True.

    Roughly equivalent to healpy.ang2pix(..., latlon=True).
    Note that the argument ordering is (ra, dec), consistent with healpy but not pixell!

    Returns pair of integer-valued arrays (idec, ira), which contain indices
    with respect to the 'shape' tuple.

    The 'ra_deg' and 'dec_deg' array arguments should have the same shape, and
    the returned idec/ira arrays will also have this shape.

    Since pixell maps can be partial-sky, pixel indices can be out of bounds.
    In this case:
    
       - If allow_outliers=False, we throw an exception.

       - If allow_outliers=True, then we "clip" the returned idec/ira arrays
         so that they are in bounds, and return a boolean mask to indicate which
         values are valid.

    Note: if you are interested in this function in order to pixelize a Catalog,
    you may want to call Catalog.to_pixell() instead.
    """

    # pixell wants a single 'coords' array, with [dec,ra] converted to radians
    coords = np.stack(np.broadcast_arrays(dec_deg, ra_deg), axis=0)   # makes copy
    coords *= (np.pi/180.)

    # pixell.enmap.sky2pix() returns a single output array, where axis 0 is {idec,ira}.
    icoords = pixell.enmap.sky2pix(
        shape,
        wcs,
        coords,
        safe = True,      # pixell does not document this argument
        corner = False,   # indicates that integer coords correspond to pixel centers (not corners)
        bcheck = (not allow_outliers)   # pixell does not document this argument
    )

    icoords = np.asarray(np.round(icoords), dtype=int)

    icoords_clipped = np.clip(
        icoords,
        a_min = 0,
        a_max = np.reshape(shape, (2,)+(1,)*(icoords.ndim-1)) - 1   # note (-1) at end
    )

    mask = np.logical_and(
        icoords[0] == icoords_clipped[0], 
        icoords[1] == icoords_clipped[1]
    )

    idec, ira = icoords_clipped   # not icoords
    
    if allow_outliers:
        return idec, ira, mask
    elif np.all(mask):
        return idec, ira

    if outlier_errmsg is None:
        outlier_errmsg = 'kszx.pixell_utils.ang2pix(): called with allow_outliers=False, and some points are out-of-bounds'
    raise RuntimeError(outlier_errmsg)


####################################################################################################


def uK_arcmin_from_ivar(ivar, max_uK_arcmin=1.0e6):
    r"""Given an inverse variance map (ivar), returns associated RMS map (in uK-arcmin).

    Based on Mat's orphics library:
       https://github.com/msyriac/orphics/blob/master/orphics/maps.py

    **Note:** currently uses pixell maps, but could easily be modified to allow healpix
    maps -- let me know if this would be useful.

    Function args:
    
      - ``ivar`` (pixell map): inverse variance map, units (uK)^{-2}.
        (For example, the return value from :func:`~kszx.act.read_ivar()`.)

      - ``max_uK_arcmin`` (float): max allowed value in the return map (prevents
        dividing by zero).

    Return value is a pixell map with units uK-arcmin.

    Equivalent to::
    
       ps = ivar.pixsizemap()
       uK_arcmin = sqrt(ps/ivar) * (60*180/np.pi)
       return np.minimum(uK_arcmin, max_uK_arcmin)

    but regulated to avoid dividing by zero.
    """

    assert isinstance(ivar, pixell.enmap.ndmap)
    assert max_uK_arcmin > 0

    a = 60*180/np.pi
    ps = ivar.pixsizemap()
    epsilon = 0.9 * np.min(ps) * (a/max_uK_arcmin)**2
    
    uK_arcmin = a * np.sqrt(ps / np.maximum(ivar,epsilon))
    return np.minimum(uK_arcmin, max_uK_arcmin)


def fkp_from_ivar(ivar, cl0, normalize=True, return_wvar=False):
    r"""Given an inverse variance map (ivar), returns associated FKP weighting.

    $$\begin{align}
    W(\theta) &= \frac{1}{C_l^{(0)} + N(\theta)} \\
    N(\theta) &\equiv \frac{\mbox{Pixel area}}{\mbox{ivar}(\theta)} \hspace{1cm} \mbox{``Local'' noise power spectrum}
    \end{align}$$

    **Note:** currently uses pixell maps, but could easily be modified to allow healpix
    maps -- let me know if this would be useful.

    Function args:

      - ``ivar`` (pixell map): inverse variance map, units (uK)^{-2}.
        (For example, the return value from :func:`~kszx.act.read_ivar()`.)

      - ``cl0`` (float): FKP weight function parameter $C_l^{(0)}$.

          - Intuitively, ``cl0`` = "fiducial signal $C_l$ at wavenumber $l$ of interest".
          - ``cl0=0`` corresponds to inverse noise weighting.
          - Large ``cl0`` corresponds to uniform weighing (but ``cl0=np.inf`` won't work).
          - I usually use ``cl0 = 0.01`` for plotting all-sky CMB temperature maps.
          - For kSZ filtering, ``cl0=3e-5`` is a reasonable choice.

      - ``normalize`` (boolean): if True, then we normalize the weight function 
        so that $\max(W(\theta))=1$.

      - ``return_wvar`` (boolean): if True, then we return $W(\theta) / ivar(\theta)$,
        instead of returning $W(\theta)$.
    
    Returns a pixell map.

    In implementation, in order to avoid divide-by-zero for ivar=0, we compute
    $W(\theta)$ equivalently as:
    
    $$W(\theta) = \frac{\mbox{ivar}(\theta)}{(\mbox{pixel area}) + C_l^{(0)} \mbox{ivar}(\theta)}$$
    """
    
    assert isinstance(ivar, pixell.enmap.ndmap)
    assert ivar.ndim == 2
    assert np.all(ivar >= 0.0)
    assert cl0 >= 0.0

    wvar = 1.0 / (ivar.pixsizemap() + cl0 * ivar)
    w = wvar * ivar
    
    wmax = np.max(w)
    assert wmax > 0.0

    ret = wvar if return_wvar else w
    
    if normalize:
        ret /= wmax

    return ret



import sys
import argparse


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='command')

    p = subparsers.add_parser('download_act')
    p.add_argument('dr', type=int, help='Either 5 or 6 (for DR5 or DR6)')

    p = subparsers.add_parser('download_desi')
    p.add_argument('survey', help='Survey name such as LRG_NGC')
    p.add_argument('-n', help='number of random files to download (default is to download all 18)')

    p = subparsers.add_parser('download_planck')

    p = subparsers.add_parser('download_sdss')
    p.add_argument('survey', help='Survey name such as CMASS_North')

    p = subparsers.add_parser('show')
    p.add_argument('filename')

    p = subparsers.add_parser('test')

    p = subparsers.add_parser('kpipe_run')
    p.add_argument('input_dirname')
    p.add_argument('output_dirname')
    p.add_argument('-n', type=int, default=400, help='number of surrogates (default 400)')
    p.add_argument('-p', type=int, default=2, help='number of processes for multiprocessing Pool (default 2))')
    
    args = parser.parse_args()

    if not hasattr(args, 'command'):
        parser.print_help()
        sys.exit(2)
    elif args.command == 'download_act':
        from . import act
        act.download(dr=args.dr)
    elif args.command == 'download_desi':
        from . import desi
        nrfiles = int(args.n) if (args.n is not None) else None
        desi.download(args.survey, dr=1, nrfiles=nrfiles)
    elif args.command == 'download_planck':
        from . import planck
        planck.download()
    elif args.command == 'download_sdss':
        from . import sdss
        sdss.download(args.survey)
    elif args.command == 'show':
        from . import io_utils
        io_utils.show_file(args.filename)
    elif args.command == 'test':
        from . import tests
        tests.run_all_tests()   # defined in kszx/tests/__init__.py
    elif args.command == 'kpipe_run':
        from . import ksz_desils
        kpipe = ksz_desils.Kpipe(args.input_dirname, args.output_dirname, nsurr=args.n)
        kpipe.run(processes=args.p)
    else:
        parser.print_help()
        sys.exit(2)



"""This source file contains some code from Selim's DESILS KSZ notebooks."""

import os
import time
import yaml
import shutil
import functools
import numpy as np
import scipy.linalg
import scipy.special
import matplotlib.pyplot as plt

from scipy.signal import savgol_filter
from scipy.ndimage import gaussian_filter1d
from scipy.interpolate import interp1d

from . import io_utils
from . import utils

from .Catalog import Catalog
from .Cosmology import Cosmology
from .KszPSE import KszPSE


####################################################################################################


class PhotozDistribution:
    def __init__(self, zobs_arr, zerr_arr, zmin=0.0, zmax=1.5, zerr_min=0.02, zerr_max=0.5, nzbins=100, nzerrbins=49, niter=100, sigma=2):
        r"""Photo-z model obtained by Richardson-Lucy deconvolution.

        Now superseded(?) by RegulatedDeconvolver, see below.
        Source: Selim's ``pipeline_getsurrogates_selim.ipynb``, Nov 2024.

        Constructor args:

          - ``zobs_arr``: 1-d array of observed redshifts.
          - ``zerr_arr``: 1-d array of observed redshift errors.
          - ``zmin, zmax, nzbins``: used internally when doing Richardson-Lucy deconvolution. 
            The defaults (0, 1.5, 100) are appropriate for DESILS-LRG.
          - ``zerr_min, zerr_max, nzerrbins``: used internally to split up the data before doing Richardson-Lucy deconvolution.
            The defaults (0.02, 0.5, 49) are approriate for DESILS-LRG.
          - ``niter``: number of Richardson-Lucy iterations.
          - ``sigma``: adjust for different smoothing levels.
        """
        
        zobs_arr = np.asarray(zobs_arr)
        zerr_arr = np.asarray(zerr_arr)
        
        # Argument checking.
        assert zobs_arr.shape == zerr_arr.shape
        assert zobs_arr.ndim == 1
        assert 0 <= zmin < zmax
        assert 0 <= zerr_min < zerr_max
        assert nzbins > 0
        assert nzerrbins > 0
        assert niter > 0
        assert sigma > 0.0

        self.nzbins = nzbins
        self.nzerrbins = nzerrbins
        self.z = np.linspace(zmin, zmax, nzbins)
        self.pzebins = np.linspace(zerr_min, zerr_max, nzerrbins+1)
        self.meanphotz = np.zeros(nzerrbins+2)
        self.umat = np.zeros((nzerrbins+2, nzbins))
        
        locatepze = np.digitize(zerr_arr, self.pzebins)

        for iz in range(nzerrbins+2):
            ix_bin = (locatepze == iz)
            zobs_bin = zobs_arr[ix_bin]
            zerr_bin = zerr_arr[ix_bin]
    
            if len(zerr_bin) < 5:
                continue
    
            self.meanphotz[iz] = np.mean(zerr_bin)
            counts_obs, bin_edges = np.histogram(zobs_bin, density=False, bins=nzbins, range=(zmin,zmax))
    
            zs_obs = (bin_edges[1:]+bin_edges[:-1])/2

            counts_obs = gaussian_filter1d(counts_obs, sigma=sigma)*1.
            counts_obs /= np.sum(counts_obs)
    
            counts_obs_int = interp1d(zs_obs, counts_obs, fill_value=0, bounds_error=False)
            observed_dist = counts_obs_int(self.z)

            Pij = np.array([[np.exp(-0.5*((zi-zj)/self.meanphotz[iz])**2) for zi in self.z] for zj in self.z])
            for j in range(len(self.z)):
                Pij[:,j] /= np.sum(Pij[:,j])

            di = observed_dist
            uguess = np.ones(nzbins)
    
            for iteration in range(niter):
                ci = np.dot(Pij, uguess)
                uguess *= np.dot(di / (ci+1.0e-10), Pij)

            self.umat[iz,:] = uguess / np.sum(uguess)

    
    def sample_z(self, zerr, zmin=0.4, zmax=1.0):
        r"""Randomly sample (ztrue, zobs) pairs, conditioned on a specified array of 'zerr' values.

        Function arguments:

          - ``zerr`` (array): 1-d array of zerr values.
          - ``(zmin, zmax)``: the returned zobs values will be constrained to satisfy 
            $z_{\rm min} < z_{\rm obs} < z_{\rm max}$. The defaults (0.4, 1.0) are appropriate
            for DESILS-LRG.

        Return values:

          - ``ztrue``: 1-d array (same length as ``zerr`` argument).
          - ``zobs``: 1-d array (same length as ``zerr`` argument).

        Each returned $(z_{\rm true}, z_{\rm obs})$ pair is consistent with the corresponding
        caller-specified $z_{\rm err}$.
        """
        
        zerr = np.asarray(zerr)
        assert zerr.ndim == 1

        locatepze = np.digitize(zerr, self.pzebins)
        ztrue = np.zeros_like(zerr)
        zobs = np.zeros_like(zerr)

        for iz in range(self.nzerrbins + 2):
            ix_bin = (locatepze == iz)
            lentruez = np.sum(ix_bin)
            
            buffer = 4
            lenpze = lentruez * buffer
    
            # Define some oversampled z-bins.
            # KMS: changed number of bins from lenpze to (10 * self.nzbins)
            # (speeds things up and gives similar results)
            
            zbin_edges = np.linspace(self.z[0], self.z[-1], 10*self.nzbins + 1)
            dz = zbin_edges[1] - zbin_edges[0]

            interp = interp1d(self.z, self.umat[iz,:])
            zvecs = (zbin_edges[1:] + zbin_edges[:-1]) / 2.  # bin centers
            zintrp = np.maximum(interp(zvecs), 0)

            # We perturb true_z within its bin [-dz/2,dz/2]
            # (Otherwise, all true_z's are at bin centers, and the histogram looks weird)
            
            true_z_sample = np.random.choice(zvecs, size=lenpze, p = (zintrp / np.sum(zintrp)))
            true_z_sample += np.random.uniform(-dz/2, dz/2, size=lenpze)
            photoz_err = np.random.normal(scale = self.meanphotz[iz], size = lenpze)
            observed_z_sample = true_z_sample + photoz_err 

            selectz_ = (observed_z_sample > zmin) * (observed_z_sample < zmax)
            ztrue[ix_bin] = (true_z_sample[selectz_])[:lentruez]
            zobs[ix_bin] = (observed_z_sample[selectz_])[:lentruez]

        return ztrue, zobs


####################################################################################################


class Kpipe:
    def __init__(self, input_dir, output_dir, nsurr=400):
        """Kendrick KSZ P(k) pipeline.

        Runnable either by calling the run() method, or from the command line with:
           python -m kszx <input_dir> <output_dir>

        Reminder: input directory must contain files { params.yml, galaxies.h5, randoms.h5,
        bounding_box.pkl }. The pipeline will populate the output directory with files
        { params.yml, pk_data.npy, pk_surrogates.npy }. (For details of these file formats,
        see one of the README files that Kendrick or Selim have lying around.)

        Note that 'class Kpipe', 'class KszPSE', and 'class CatalogGridder' are related
        as follows:

          - Kpipe: high-level pipeline, knows about file inputs/outputs, defers
              heavy lifting to 'class KszPSE'.

          - KszPSE: KSZ power spectrum estimation and surrogate generation logic
              is here.

          - CatalogGridder: a lower-level class which supplies normalizations (both
              field-level and power spectrum level).
        """
        
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.nsurr = nsurr
        
        # Create output_dir and 'tmp' subdir
        os.makedirs(f'{output_dir}/tmp', exist_ok=True)

        with open(f'{input_dir}/params.yml', 'r') as f:
            params = yaml.safe_load(f)

            self.surr_bg = params['surr_bg']
            self.surr_ic_nbins = params['surr_ic_nbins']
            self.spin0_hack = params['spin0_hack']
            assert self.spin0_hack in [ True, False ]

            self.kmax = params['kmax']
            self.nkbins = params['nkbins']
            self.kbin_edges = np.linspace(0, self.kmax, self.nkbins+1)
            self.kbin_centers = (self.kbin_edges[1:] + self.kbin_edges[:-1]) / 2.

        self.box = io_utils.read_pickle(f'{input_dir}/bounding_box.pkl')

        # This code moved into cached properties (see below), in order to reduce startup time.
        # self.cosmo = Cosmology('planck18+bao')
        # self.gcat = Catalog.from_h5(f'{input_dir}/galaxies.h5')        
        # self.rcat = Catalog.from_h5(f'{input_dir}/randoms.h5')

        self.pk_data_filename = f'{output_dir}/pk_data.npy'
        self.pk_surr_filename = f'{output_dir}/pk_surrogates.npy'
        self.pk_single_surr_filenames = [ f'{output_dir}/tmp/pk_surr_{i}.npy' for i in range(nsurr) ]


    @functools.cached_property
    def cosmo(self):
        return Cosmology('planck18+bao')

    @functools.cached_property
    def gcat(self):
        return Catalog.from_h5(f'{self.input_dir}/galaxies.h5')

    @functools.cached_property
    def rcat(self):
        return Catalog.from_h5(f'{self.input_dir}/randoms.h5')

    
    @functools.cached_property
    def pse(self):
        """Returns a KszPSE object."""
        
        print('Initializing KszPSE: this will take a few minutes')
        
        # FIXME needs comment
        surr_ngal_mean = self.gcat.size
        surr_ngal_rms = 4 * np.sqrt(self.gcat.size)  # 4x Poisson
        
        rweights = getattr(self.rcat, 'weight_zerr', None)
        vweights = getattr(self.rcat, 'vweight_zerr', None)
        
        pse = KszPSE(
            box = self.box, 
            cosmo = self.cosmo, 
            randcat = self.rcat, 
            kbin_edges = self.kbin_edges,
            surr_ngal_mean = surr_ngal_mean,
            surr_ngal_rms = surr_ngal_rms,
            surr_bg = self.surr_bg,
            rweights = rweights,
            nksz = 2,
            ksz_rweights = vweights,
            ksz_bv = [ self.rcat.bv_90, self.rcat.bv_150 ],
            ksz_tcmb_realization = [ self.rcat.tcmb_90, self.rcat.tcmb_150 ],
            ztrue_col = 'ztrue',
            zobs_col = 'zobs',
            surr_ic_nbins = self.surr_ic_nbins,
            spin0_hack = self.spin0_hack
        )
        
        print('KszPSE initialization done')
        return pse

    
    def get_pk_data(self, run=False):
        """Returns a shape (3,3,nkbins) array.

        If run=False, then this function expects the P(k) file to be on disk from a previous pipeline run.
        If run=True, then the P(k) file will be computed if it is not on disk.
        """

        if os.path.exists(self.pk_data_filename):
            return io_utils.read_npy(self.pk_data_filename)
        
        if not run:
            raise RuntimeError(f'Kpipe.get_pk_data(): run=False was specified, and file {self.pk_data_filename} not found')

        print('get_pk_data(): running\n', end='')
        
        t90 = utils.subtract_binned_means(self.gcat.tcmb_90, self.gcat.z, nbins=25)
        t150 = utils.subtract_binned_means(self.gcat.tcmb_150, self.gcat.z, nbins=25)

        gweights = getattr(self.gcat, 'weight_zerr', None)
        vweights = getattr(self.gcat, 'vweight_zerr', None)

        pk_data = self.pse.eval_pk(
            gcat = self.gcat,
            gweights = gweights,
            ksz_gweights = vweights,
            ksz_bv = [ self.gcat.bv_90, self.gcat.bv_150 ], 
            ksz_tcmb = [ t90, t150 ],
            zobs_col = 'z'
        )

        io_utils.write_npy(self.pk_data_filename, pk_data)
        return pk_data


    def get_pk_surrogate(self, isurr, run=False):
        """Returns a shape (6,6,nkbins) array.
        
        If run=False, then this function expects the P(k) file to be on disk from a previous pipeline run.
        If run=True, then the P(k) file will be computed if it is not on disk.
        """
        
        fname = self.pk_single_surr_filenames[isurr]
        
        if os.path.exists(fname):
            return io_utils.read_npy(fname)

        if not run:
            raise RuntimeError(f'Kpipe.get_pk_surrogate(): run=False was specified, and file {fname} not found')

        print(f'get_pk_surrogate({isurr}): running\n', end='')
        
        self.pse.simulate_surrogate()
        
        for sv in [ self.pse.Sv_noise, self.pse.Sv_signal ]:
            assert sv.shape ==  (2, self.rcat.size)
            for j in range(2):
                sv[j,:] = utils.subtract_binned_means(sv[j,:], self.rcat.zobs, nbins=25)
    
        pk = self.pse.eval_pk_surrogate()

        io_utils.write_npy(fname, pk)
        return pk


    def get_pk_surrogates(self):
        """Returns a shape (nsurr,6,6,nkins) array, containing P(k) for all surrogates."""
        
        if os.path.exists(self.pk_surr_filename):
            return io_utils.read_npy(self.pk_surr_filename)

        if not all(os.path.exists(f) for f in self.pk_single_surr_filenames):
            raise RuntimeError(f'Kpipe.read_pk_surrogates(): necessary files do not exist; you need to call Kpipe.run()')

        pk = np.array([ io_utils.read_npy(f) for f in self.pk_single_surr_filenames ])
        
        io_utils.write_npy(self.pk_surr_filename, pk)
        return pk


    def run(self, processes=2):
        """Runs pipeline. If any output files already exist, they will be skipped."""
        
        # Copy yaml file from input to output dir.
        if not os.path.exists(f'{self.output_dir}/params.yml'):
            shutil.copyfile(f'{self.input_dir}/params.yml', f'{self.output_dir}/params.yml')
                
        have_data = os.path.exists(self.pk_data_filename)
        have_surr = os.path.exists(self.pk_surr_filename)
        missing_surrs = [ ] if have_surr else [ i for (i,f) in enumerate(self.pk_single_surr_filenames) if not os.path.exists(f) ]

        if (not have_surr) and (len(missing_surrs) == 0):
            self.get_pk_surrogates()   # creates "top-level" file
            have_surr = True
            
        if have_data and have_surr:
            print(f'Kpipe.run(): pipeline has already been run, exiting early')
            return
        
        # Initialize KszPSE before creating multiprocessing Pool.
        self.pse

        # FIXME currently I don't have a good way of setting the number of processes automatically --
        # caller must adjust the number of processes to the amount of memory in the node.
        
        with utils.Pool(processes) as pool:
            l = [ ]
                
            if not have_data:
                l += [ pool.apply_async(self.get_pk_data, (True,)) ]
            for i in missing_surrs:
                l += [ pool.apply_async(self.get_pk_surrogate, (i,True)) ]

            for x in l:
                x.get()

        if not have_surr:
            # Consolidates all surrogates into one file
            self.get_pk_surrogates()


####################################################################################################


class RegulatedDeconvolver:
    def __init__(self, zobs_vec, zerr_vec, zbin_width, soft_zmax=None):
        """A photo-z error deconvolver which avoids the oscillatory behavior of Lucy-Richardson.
        
        Given the observed 2-d (zobs,zerr) distribution, models the 3-d (ztrue,zobs,zerr) distribution.
        FIXME: needs comments explaining how it works!
        
        - zobs_vec: array of shape (ngal,) containing observed (photometric) redshifts.

        - zerr_vec: array of shape (ngal,) containing estimated photo-z errors.

        - zbin_width (scalar): used internally for binning. (Recommend ~0.01 for DESILS-LRG.)

        - soft_zmax (scalar): used internally; galaxies with (zobs > soft_zmax) are discarded.
          Should be chosen significantly larger than the max redshift of interest.
        
          (For DESILS-LRG with zmax=1, recommend soft_zmax=1.5. If this parameter is omitted,
          then you'll get a tail of rare galaxies with z~5, which is annoying.)
        """
        
        assert zobs_vec.ndim == 1
        assert zobs_vec.shape == zerr_vec.shape
        assert np.all(zobs_vec >= 0)
        assert np.all(zerr_vec > 0)

        if soft_zmax is not None:
            mask = (zobs_vec <= soft_zmax)
            zobs_vec = zobs_vec[mask]
            zerr_vec = zerr_vec[mask]
        
        self.zobs_vec = zobs_vec
        self.zerr_vec = zerr_vec
        self.zbin_width = zbin_width
        self.zmin = np.min(zobs_vec)
        self.zmax = np.max(zobs_vec)
        self.soft_zmax = soft_zmax

        self.nzbins = np.array(self.zmax/zbin_width, dtype=int)
        self.zbin_edges = np.arange(self.nzbins+1) * zbin_width
        
        # The variable y = log(zerr + zbin_width) is used to define zerr bins.
        y = np.log(zerr_vec + self.zbin_width)
        self.ymin = np.min(y) - 1.0e-8
        self.ymax = np.max(y) + 1.0e-8
        self.nybins = int((self.ymax - self.ymin) / 0.2) + 1
        self.ybin_width = (self.ymax - self.ymin) / self.nybins
        
        iz = np.array(zobs_vec / zbin_width, dtype=int)
        iz = np.clip(iz, 0, self.nzbins-1)

        iy = np.array((y - self.ymin) / self.ybin_width, dtype=int)
        iy = np.clip(iy, 0, self.nybins-1)

        # self.pobs[iy,iz] is an array of shape (self.nybins, self.nzbins).
        self.pobs = np.bincount(iy*self.nzbins + iz, minlength=self.nybins*self.nzbins)
        self.pobs = np.reshape(self.pobs, (self.nybins, self.nzbins))
        self.pobs += 1   # regulate
        self.pobs = self.pobs / np.sum(self.pobs,axis=1).reshape((-1,1))  # normalize to PDF in z-bins (for each y-bin)


    def _sample(self, zobs, zerr):
        """Helper for sample(): Sample from conditional distribution P(ztrue|zobs,zerr)."""
        
        n = len(zobs)
        nb = self.nzbins
        rng = np.random.default_rng()

        # iy, iz = bin indices (shape (n,))
        # Reminder: the variable y = log(zerr + zbin_width) is used to define zerr bins.
        y = np.log(zerr + self.zbin_width)
        iy = np.array((y - self.ymin) / self.ybin_width, dtype=int)
        iy = np.clip(iy, 0, self.nybins-1)
        
        # pobs = shape(n,nb)
        pobs = self.pobs[iy]
        assert pobs.shape == (n,nb)

        # cdf = shape (n,nb+1), not "reweighted by pobs".
        cdf = self.C((self.zbin_edges.reshape((1,-1)) - zobs.reshape((-1,1))) / zerr.reshape((-1,1)))
        assert cdf.shape == (n,nb+1)
        
        # pdf = shape (n,nb), "reweighted by pobs"
        pdf = cdf[:,1:] - cdf[:,:-1]
        pdf *= pobs
        pdf /= np.sum(pdf,axis=1).reshape((-1,1))

        # ztrue_bins has shape (n,). I don't think there's any way to vectorize this loop.
        ztrue_bins = np.array([ rng.choice(nb, p=pdf[i,:]) for i in range(n) ])
        
        cdf_lo = cdf[np.arange(n), ztrue_bins]
        cdf_hi = cdf[np.arange(n), ztrue_bins+1]
        c = cdf_lo + (cdf_hi-cdf_lo) * rng.uniform(size=n)
        ztrue = zobs + zerr * self.Cinv(c)

        return ztrue

        
    def sample(self, n, zobs_min=None, zobs_max=None):
        """Returns (ztrue, zobs, zerr), where all 3 arrays are shape (n,).'
        If n is large, you may want to call sample_parallel() instead."""

        zobs_min = zobs_min if (zobs_min is not None) else (self.zmin - 0.01)
        zobs_max = zobs_max if (zobs_max is not None) else (self.zmax + 0.01)
        mask = np.logical_and(self.zobs_vec >= zobs_min, self.zobs_vec <= zobs_max)
        rng = np.random.default_rng()
        
        # Shape (n,)
        ix = np.nonzero(mask)[0]
        ix = rng.choice(ix, size=n, replace=True)
        zobs = self.zobs_vec[ix]
        zerr = self.zerr_vec[ix]
        ztrue = np.zeros(n)
        pos = 0

        # Compute in batches, to keep memory usage under control.
        while pos < n:
            pos2 = min(n, pos + 10**5)
            ztrue[pos:pos2] = self._sample(zobs[pos:pos2], zerr[pos:pos2])
            pos = pos2
                
        return ztrue, zobs, zerr


    def sample_parallel(self, n, zobs_min=None, zobs_max=None, processes=None):
        """Returns (ztrue, zobs, zerr), where all 3 arrays are shape (n,).
        
        Uses a multiprocessing Pool for speed. If 'processes' is None, then a
        sensible default will be chosen."""

        with utils.Pool(processes) as pool:
            p = pool._processes
            nlist = [ (((i+1)*n)//p) - ((i*n)//p) for i in range(p) ]
            m = pool.starmap_async(self.sample, [(n,zobs_min,zobs_max) for n in nlist])
            result = m.get()  # list of (ztrue, zobs, zerr) triples
            
        # Concatenate results from each task.
        ztrue = np.concatenate([r[0] for r in result])
        zobs = np.concatenate([r[1] for r in result])
        zerr = np.concatenate([r[2] for r in result])

        return ztrue, zobs, zerr


    @classmethod
    def C(cls, x):
        return (1/2) * scipy.special.erf(x / np.sqrt(2.))

    @classmethod
    def Cinv(cls, y):
        return scipy.special.erfinv(2*y) * np.sqrt(2.)


####################################################################################################


class PipelineOutdir:
    def __init__(self, dirname, nsurr=None):
        """
        PipelineOutdir: a helper class for postprocessing/plotting pipeline outputs.        
        (Note: there is a separate class 'PgvLikelihood' for MCMCs and parameter fits, see below)

        The constructor reads the following files::
        
          {dirname}/params.yml         # for nkbins, kmax
          {dirname}/pk_data.npy        # shape (3, 3, nkbins)
          {dirname}/pk_surrogates.npy  # shape (nsurr, 6, 6, nkbins)
          
        Constructor arguments:

          - dirname (string): name of pipeline output directory.
         
          - nsurr (integer or None): this is a hack for running on an incomplete
            pipeline. If specified, then {dirname}/pk_surr.npy is not read.
            Instead we read files of the form {dirname}/tmp/pk_surr_{i}.npy.
        """

        filename = f'{dirname}/params.yml'
        print(f'Reading {filename}')
        
        with open(filename, 'r') as f:
            params = yaml.safe_load(f)

        kmax = params['kmax']
        nkbins = params['nkbins']
        surr_bg = params['surr_bg']
        kbin_edges = np.linspace(0, kmax, nkbins+1)
        kbin_centers = (kbin_edges[1:] + kbin_edges[:-1]) / 2.

        pk_data = io_utils.read_npy(f'{dirname}/pk_data.npy')
        if pk_data.shape != (3,3,nkbins):
            raise RuntimeError(f'Got {pk_data.shape=}, expected (3,3,nkbins) where {nkbins=}')

        if nsurr is None:
            pk_surr = io_utils.read_npy(f'{dirname}/pk_surrogates.npy')
            if (pk_surr.ndim != 4) or (pk_surr.shape[1:] != (6,6,nkbins)):
                raise RuntimeError(f'Got {pk_surr.shape=}, expected (nsurr,6,6,nkbins) where {nkbins=}')

        else:
            assert nsurr >= 1
            pk_surr = [ ]
            
            for i in range(nsurr):
                pk_surrogate = io_utils.read_npy(f'{dirname}/tmp/pk_surr_{i}.npy')
                if pk_surrogate.shape != (6,6,nkbins):
                    raise RuntimeError(f'Got {pk_surrogate.shape=}, expected (6,6,nkbins) where {nkbins=}')
                pk_surr.append(pk_surrogate)
                
            pk_surr = np.array(pk_surr)

        self.k = kbin_centers
        self.nkbins = nkbins
        self.kmax = kmax
        self.dk = kmax / nkbins
        
        self.pk_data = pk_data
        self.pk_surr = np.array(pk_surr)
        self.nsurr = self.pk_surr.shape[0]
        self.surr_bg = surr_bg


    def pgg_data(self):
        """Returns shape (nkbins,) array containing P_{gg}^{data}."""
        return self.pk_data[0,0,:]
        
    def pgg_mean(self, fnl=0):
        """Returns shape (nkbins,) array, containing <P_{gg}^{surr}>."""
        return np.mean(self._pgg_surr(fnl), axis=0)

    def pgg_rms(self, fnl=0):
        """Returns shape (nkbins,) array, containing sqrt(Var(P_{gg}^{surr}))."""
        return np.sqrt(np.var(self._pgg_surr(fnl), axis=0))

    
    def pgv_data(self, field):
        """Returns shape (nkbins,) array containing P_{gv}^{data}.

        The 'field' argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - field=[1,0] for 90 GHz
           - field=[0,1] for 150 GHz
           - field=[1,-1] for null (90-150) GHz.
        """
        field = self._check_field(field)
        return field[0]*self.pk_data[0,1] + field[1]*self.pk_data[0,2]
        
    def pgv_mean(self, field, fnl, bv):
        """Returns shape (nkbins,) array containing <P_{gv}^{surr}>.

        The 'field' argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - field=[1,0] for 90 GHz
           - field=[0,1] for 150 GHz
           - field=[1,-1] for null (90-150) GHz.
        """
        return np.mean(self._pgv_surr(field,fnl,bv), axis=0)

    def pgv_rms(self, field, fnl, bv):
        """Returns shape (nkbins,) array containing sqrt(Var(P_{gv}^{surr})).

        The 'field' argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - field=[1,0] for 90 GHz
           - field=[0,1] for 150 GHz
           - field=[1,-1] for null (90-150) GHz.
        """
        return np.sqrt(np.var(self._pgv_surr(field,fnl,bv), axis=0))


    def pvv_data(self, field):
        """Returns shape (nkbins,) array containing P_{vv}^{data}.

        The 'field' argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - field=[1,0] for 90 GHz
           - field=[0,1] for 150 GHz
           - field=[1,-1] for null (90-150) GHz.
        """
        field = self._check_field(field)
        t = field[0]*self.pk_data[1,:] + field[1]*self.pk_data[2,:]  # shape (3,)
        return field[0]*t[1] + field[1]*t[2]
        
    def pvv_mean(self, field, bv):
        """Returns shape (nkbins,) array containing < P_{vv}^{data} >.

        The 'field' argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - field=[1,0] for 90 GHz
           - field=[0,1] for 150 GHz
           - field=[1,-1] for null (90-150) GHz.
        """
        return np.mean(self._pvv_surr(field,bv), axis=0)
        
    def pvv_rms(self, field, bv):
        """Returns shape (nkbins,) array containing sqrt(var(P_{vv}^{data})).

        The 'field' argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - field=[1,0] for 90 GHz
           - field=[0,1] for 150 GHz
           - field=[1,-1] for null (90-150) GHz.
        """        
        return np.sqrt(np.var(self._pvv_surr(field,bv), axis=0))


    def _check_field(self, field):
        """Checks that 'field' is a 1-d array of length 2."""
        field = np.array(field, dtype=float)
        if field.shape != (2,):
            raise RuntimeError(f"Expected 'field' argument to be a 1-d array of length 2, got {field.shape=}")
        return field
    
    def _pgg_surr(self, fnl=0):
        """Returns shape (nsurr, nkbins) array, containing P_{gg} for each surrogate"""
        pgg = self.pk_surr[:,:2,:2,:]             # shape (nsurr, 2, 2, nkbins)
        pgg = pgg[:,0,:,:] + fnl * pgg[:,1,:,:]   # shape (nsurr, 2, nkbins)
        return pgg[:,0,:] + fnl * pgg[:,1,:]      # shape (nsurr, nkbins)
    
    def _pgv_surr(self, field, fnl, bv):
        """Returns shape (nsurr, nkbins) array, containing P_{gv} for each surrogate"""
        field = self._check_field(field)
        pgv = self.pk_surr[:,:2,2:,:]                                # shape (nsurr, 2, 4, nkbins)
        pgv = pgv[:,0,:,:] + fnl * pgv[:,1,:,:]                      # shape (nsurr, 4, nkbins)
        pgv = field[0] * pgv[:,0:2,:] + field[1] * pgv[:,2:4,:]      # shape (nsurr, 2, nkbins)
        return pgv[:,0,:] + bv * pgv[:,1,:]                          # shape (nsurr, nkbins)

    def _pvv_surr(self, field, bv):
        """Returns shape (nsurr, nkbins) array, containing P_{vv} for each surrogate"""
        field = self._check_field(field)
        pvv = self.pk_surr[:,2:,2:,:]                                  # shape (nsurr, 4, 4, nkbins)
        pvv = field[0] * pvv[:,0:2,:,:] + field[1] * pvv[:,2:4,:,:]    # shape (nsurr, 2, 4, nkbins)
        pvv = field[0] * pvv[:,:,0:2,:] + field[1] * pvv[:,:,2:4,:]    # shape (nsurr, 2, 2, nkbins)
        pvv = pvv[:,0,:,:] + bv * pvv[:,1,:,:]                         # shape (nsurr, 2, nkbins)
        return pvv[:,0,:] + bv * pvv[:,1,:]                            # shape (nsurr, nkbins)


####################################################################################################


class PgvLikelihood:
    def __init__(self, data, surrs, bias_matrix, jeffreys_prior=False):
        """Likelihood function for one or more power spectra of the form P_{gv}(k).

        The PgvLikelihood constructor has an abstract syntax, which may not be the
        most convenient. Instead of calling the constructor directly, you may want
        to call the from_pipeline_outdir() static method as follows::

           dirname = ...   # directory name containing pipeline outputs
           pout = PipelineOutputs(dirname)

           # The 'fields=[[1,0]]' argument selects 90 GHz.
           # See the from_pipeline_outdir() docstring for more examples.
           lik = PgvLikelihood.from_pipeline_outdir(pipeline_outputs, fields=[[1,0]], nkbins=10)

        The rest of this docstring describes the PgvLikelihood constructor syntax
        (even though, as explained above, you probably don't want to call the constructor
        directly!) First we define::

          B = number of bias parameters in the MCMC (usually 1)
          V = number of velocity reconstructions (usually 1)
          K = number of k-bins
          S = number of surrogate sims
          D = V*K = number of components in "data vector".

        Note the distinction between a 90+150 GHz "sum" fit (V=1), and a 90+150 GHz
        "joint" fit (V=2). In the first case, we add the 90+150 bandpowers and construct
        a likelihood based on their sum, and in the second case we do a joint fit to
        both sets of bandpowers (i.e. doubling the size of the data vector and
        covariance matrices).
        
        Then the constructor arguments are as follows:

          - data: P_{gv} "data" bandpowers, as array of shape (V,K).
        
          - surrs: P_{gv} surrogate bandpowers, as array of shape (S,2,2,V,K).
              where the first length-2 index is fnl exponent {0,1}
              and the second length-2 index is bias exponent {0,1}

          - bias_matrix: array of shape (B,V), where B <= V. This gives the
            correspondence between bias parameters and velocity reconstruction.
            There are basically 3 cases of interest here:

              - single field (V=1):
                  bias_matrix=[[1]]
        
              - joint analysis (V=2), bias assumed to be the same for both freqs:
                  bias_matrix = [[1,1]]

              - joint analysis (V=2) with two independent bias parameters:
                  bias_matrix = [[1,0],[0,1]]
        """

        data = np.asarray(data, dtype=float)
        surrs = np.asarray(surrs, dtype=float)
        bias_matrix = np.asarray(bias_matrix, dtype=float)
        
        if data.ndim != 2:
            raise RuntimeError(f'got {data.shape=}, expected (V,K)')
        if (surrs.ndim != 5) or (surrs.shape[1:3] != (2,2)):
            raise RuntimeError(f'got {surrs.shape=}, expected (S,2,2,V,K)')
        if bias_matrix.ndim != 2:
            raise RuntimeError(f'got {bias_matrix.shape=}, expected (B,V)')

        if data.shape != surrs.shape[3:]:
            raise RuntimeError(f'data/surrs have inconsistent shapes (expected (V,K) and (S,2,2,V,K), got {data.shape} and {surr.shape})')
        if data.shape[0] != bias_matrix.shape[1]:
            raise RuntimeError(f'data/bias_matrix have inconsistent shapes (expected (V,K) and (B,V), got {data.shape} and {bias_matrix.shape})')
        if bias_matrix.shape[0] > bias_matrix.shape[1]:
            raise RuntimeError(f'expected B <= V, got (B,V)={bias_matrix.shape}')

        self.B = bias_matrix.shape[0]
        self.V = bias_matrix.shape[1]
        self.K = data.shape[1]
        self.S = surrs.shape[0]
        self.D = self.V * self.K
        
        self.data = data                                 # shape (V,K)
        self.data_vector = np.reshape(data, (self.D,))   # shape (D,)
        self.surrs = surrs
        self.bias_matrix = bias_matrix
        self.jeffreys_prior = jeffreys_prior

        self._init_fast_likelihood()


    @staticmethod
    def from_pipeline_outdir(pout, fields, nkbins, multi_bias=None, jeffreys_prior=None):
        """
        Constructs a PgvLikelhood from a PipelineOutput object. Usually more convenient
        than calling the PgvLikeihood constructor directly.

        The 'pout' argument is a PipelineOutdir object.
        
        The 'fields' argument is a V-by-2 matrix. Each row of the matrix selects a linear
        combination of the 90+150 GHz velocity reconstructions. For example:
        
           - fields = [[1,0]] for 90 GHz analysis
           - fields = [[0,1]] for 150 GHz analysis
           - fields = [[1,1]] for (90+150) "sum map" analysis
           - fields = [[1,-1]] for null (90-150) "null map" analysis
           - fields = [[1,0],[0,1]] for joint analysis with both (90+150 GHz) sets of bandpowers

        The 'multi_bias' argument only needed for a joint analysis (i.e. V > 1) and determines
        whether each freq channel has an independent bias parameter (multi_bias=True), or whether
        all frequency channels have the same bias.
        """
        
        assert isinstance(pout, PipelineOutdir)
        assert 4 <= nkbins <= pout.nkbins
        
        fields = np.array(fields)
        if (fields.ndim != 2) or (fields.shape[0] < 1) or (fields.shape[1] != 2):
            raise RuntimeError(f"PgvLikelihood: expected 'fields' arg to have shape (V,2), got shape {fields.shape}")

        K = nkbins
        S = pout.nsurr
        V = fields.shape[0]

        if (V > 1) and (multi_bias is None):
            raise RuntimeError(f"The 'multi_bias' argument is required if nfields > 1.")

        # PgvLikelihood constructor expects data array of shape (V,K)
        pgv_data = pout.pk_data[0,1:3,:K]      # shape (2,K)
        pgv_data = [ (f[0]*pgv_data[0,:] +f[1]*pgv_data[1,:]) for f in fields ]  # shape (V,K)

        # PgvLikelihood constructor expects surrogate array of shape (S,2,2,V,K).
        pgv_surr = pout.pk_surr[:,0:2,2:6,:K]           # shape (S,2,4,K)
        pgv_surr = np.reshape(pgv_surr, (S,2,2,2,K))    # shape (S,2,2,2,K), length-2 indices are (fnl_exponent, freq, bv_exponent)
        pgv_surr = [ (f[0]*pgv_surr[:,:,0,:,:] + f[1]*pgv_surr[:,:,1,:,:]) for f in fields ]  # shape (V,S,2,2,K)
        pgv_surr = np.transpose(pgv_surr, (1,2,3,0,4))  # shape (S,2,2,V,K)

        bias_matrix = np.identity(V) if multi_bias else np.ones((1,V))
        return PgvLikelihood(pgv_data, pgv_surr, bias_matrix, jeffreys_prior)

    
    def specialize_surrogates(self, fnl, bv, flatten):
        """
        Returns a shape (S,V,K) array, by "specializing" surrogates to (fnl, bv).
        Convenient but slow! Used in many places, but not fast_likelihood().
        """

        S, K, V, D = self.S, self.K, self.V, self.D
        
        fnl = float(fnl)        
        b = self._validate_bv(bv)        # shape (B,)
        b = np.dot(b, self.bias_matrix)  # shape (V,)
        b = np.reshape(b, (1,V,1))       # shape (1,V,1)

        # Apply fnl, obtaining shape (S,2,V,K)
        s = self.surrs[:,0,:,:,:] + fnl * self.surrs[:,1,:,:,:]

        # Apply bv, obtaining shape (S,V,K)
        s = s[:,0,:,:] + b * s[:,1,:,:]

        # Return either shape (S,V,K) or shape (S,D), depending on whether flatten=True.
        return np.reshape(s,(S,D)) if flatten else s
                
        
    def slow_mean_and_cov(self, fnl, bv):
        """Returns (mean, cov) where mean.shape=(D,) and cov.shape=(D,D)."""
        
        s = self.specialize_surrogates(fnl, bv, flatten=True)  # shape (S,D)
        mean = np.mean(s, axis=0)       # shape (D,)
        cov = np.cov(s, rowvar=False)   # shape (D, D)
        return mean, cov

    
    def slow_mean_and_cov_gradients(self, fnl, bv):
        """Returns grad(mu), grad(cov). Used to compute Jeffreys prior.

        Both gradients are with respect to (fnl,bv) and are represented as arrays 
        with an extra length-(B+1) axis, i.e.

          mu.shape = (D,)    =>  grad_mu.shape = (B+1,D)
          cov.shape = (D,D)  =>  grad_cov.shape = (B+1,D,D)
          
        Uses boneheaded algorithm: since mu, C are at most quadratic in fNL (and bv),
        naive finite difference is exact (and independent of step sizes).
        """

        B, D = self.B, self.D

        fnl = float(fnl)        
        bv = self._validate_bv(bv)   # shape (B,)

        # Parameter vectors of shape (B+1,)
        x0 = np.concatenate(((fnl,), bv))
        dx = np.concatenate(((50,), np.full(B,0.1)))
        
        mu_p = np.zeros((B+1, D))
        mu_n = np.zeros((B+1, D))
        cov_p = np.zeros((B+1, D, D))
        cov_n = np.zeros((B+1, D, D))

        for i in range(B+1):
            xp = np.copy(x0)
            xn = np.copy(x0)
            xp[i] += dx[i]
            xn[i] -= dx[i]

            mu_p[i,:], cov_p[i,:,:] = self.slow_mean_and_cov(xp[0], xp[1:])
            mu_n[i,:], cov_n[i,:,:] = self.slow_mean_and_cov(xn[0], xn[1:])
   
        grad_mu = (mu_p - mu_n) / (2 * dx.reshape((B+1,1)))
        grad_cov = (cov_p - cov_n) / (2 * dx.reshape((B+1,1,1)))
 
        return grad_mu, grad_cov
        
        
    def slow_log_likelihood(self, fnl, bv):
        """Returns the scalar quantity log(L)."""
        
        mean, cov = self.slow_mean_and_cov(fnl, bv)

        x = self.data_vector - mean
        cinv = np.linalg.inv(cov)
        sign, logabsdet = np.linalg.slogdet(cov)
        assert sign == 1

        # log L = -(1/2) log(det C) - (1/2) x^T C^{-1} x
        logL = -0.5 * np.dot(x, np.dot(cinv,x))
        logL -= 0.5 * logabsdet

        if self.jeffreys_prior:
            grad_mu, grad_cov = self.slow_mean_and_cov_gradients(fnl, bv)
            cinv_dmu = [ np.dot(cinv,dmu) for dmu in grad_mu ]
            cinv_dc = [ np.dot(cinv,dc) for dc in grad_cov ]

            # (B+1)-by-(B+1) Fisher matrix            
            B = self.B
            f = np.zeros((B+1,B+1))
            for i in range(B+1):
                for j in range(B+1):
                    f[i,j] += np.dot(grad_mu[i], cinv_dmu[j])  # no factor (1/2)
                    f[i,j] += 0.5 * np.trace(np.dot(cinv_dc[i], cinv_dc[j]))

            # Jeffreys prior is equivalent to including sqrt(det(F)) in the likelihood.
            sign, logabsdet_F = np.linalg.slogdet(f)
            logL += 0.5 * logabsdet_F
            assert sign == 1
            
        return logL

    
    def run_mcmc(self, nwalkers=8, nsamples=10000, discard=1000, thin=5):
        """
        Initializes self.samples to an array of shape (N,2) where N is large,
        and the length-2 axis represents {fnl,bv}.
        """

        import emcee
        print(f'MCMC start: {nwalkers=}, {nsamples=}, {discard=}, {thin=}')

        x0 = np.zeros((nwalkers, self.B+1))
        x0[:,0] = np.random.uniform(-50, 50, size=nwalkers)  # fnl
        x0[:,1:] = np.random.uniform(0.5, 1.0, size=(nwalkers,self.B))  # bv

        logL = lambda x: self.fast_log_likelihood(x[0], x[1:])
        sampler = emcee.EnsembleSampler(nwalkers, self.B+1, logL)
        sampler.run_mcmc(x0, nsamples)
        self.samples = sampler.get_chain(discard=discard, thin=thin, flat=True)
        
        print('MCMC done. To see the results, call the show_mcmc() method.')


    def show_mcmc(self, title=None):
        """Makes a corner plot from MCMC results."""

        if not hasattr(self, 'samples'):
            raise RuntimeError('Must call ')
        
        import corner
        
        fig = corner.corner(self.samples, bins=100, range=(0.99,0.99), labels=[r'$f_{NL}$',r'$b_v$'])
        if title is not None:
            fig.suptitle(title)
        
        plt.show()

        fnl_samples = self.samples[:,0]  # shape (nsamp,)
        bv_samples = self.samples[:,1:]  # shape (nsamp,B)
        qlevels = [ 0.025, 0.16, 0.5, 0.84, 0.975 ]
        
        fnl_quantiles = np.quantile(fnl_samples, qlevels)
        for q,fnl in zip(qlevels, fnl_quantiles):
            s = f'  ({(fnl-fnl_quantiles[2]):+.03f})' if (q != 0.5) else ''
            print(f'{(100*q):.03f}% quantile: {fnl=:.03f}{s}')

        for b in range(self.B):
            bv_quantiles = np.quantile(bv_samples[:,b], qlevels)
            print(f'\nBias parameter {b}')
            for q,bv in zip(qlevels, bv_quantiles):
                s = f'  ({(bv-bv_quantiles[2]):+.03f})' if (q != 0.5) else ''
                print(f'{(100*q):.03f}% quantile: {bv=:.03f}{s}')

        print(f'\nSNR: {self.compute_snr():.03f}')


    def analyze_chi2(self, fnl, bv, ddof=None):
        """Do model parameters (fnl,bv) fit the data? Returns (chi2, ndof, p-value).

        The 'ddof' argument is used to compute the number of degrees of freedom:
            ndof = nkbins - ddof

        If ddof=None, then it will be equal to the number of nonzero (fnl, bias) params.
        """
        
        fnl = float(fnl)        
        bv = self._validate_bv(bv)   # shape (B,)
        mean, cov = self.slow_mean_and_cov(fnl, bv)

        if ddof is None:
            ddof = 1 if (fnl != 0) else 0
            ddof += np.count_nonzero(bv)
        
        x = self.data_vector - mean
        chi2 = np.dot(x, np.linalg.solve(cov,x))
        ndof = self.K - ddof
        pte = scipy.stats.chi2.sf(chi2, ndof)
        
        return chi2, ndof, pte
        
    
    def fit_fnl_and_bv(self, fnl0=0, bv0=0.3):
        """Returns (fnl, bv) obtained by maximizing joint likelihood."""

        x0 = np.zeros(self.B+1)
        x0[0] = fnl0
        x0[1:] = bv0

        f = lambda x: -self.fast_log_likelihood(x[0], x[1:])  # note minus sign
        result = scipy.optimize.minimize(f, x0, method='Nelder-Mead')
        
        fnl = result.x[0]
        bv = result.x[1:]
        return fnl, bv        

        
    def fit_bv(self, fnl=0, bv0=0.3):
        """Returns bv obtained by maximizing conditional likelihood at the given fNL."""

        x0 = np.full(self.B, bv0)
        f = lambda x: -self.fast_log_likelihood(fnl, x)  # note minus sign
        result = scipy.optimize.minimize(f, x0, method='Nelder-Mead')
        
        bv = result.x
        return bv


    def compute_snr(self):
        # This implementation works even if there are multiple bias params (B > 1).
        B, D = self.B, self.D
        
        _, cov = self.slow_mean_and_cov(0, np.zeros(self.B))                # discard mean
        grad_mu, _ = self.slow_mean_and_cov_gradients(0, np.zeros(self.B))  # discard grad_cov
        m = grad_mu[1:,:]           # shape (B,D)       
        d = self.data_vector        # shape (D,)

        cinv_m = np.linalg.solve(cov, m.T)  # shape (D,B)
        h = np.dot(m, cinv_m)               # shape (B,B)
        g = np.dot(d, cinv_m)

        dchisq = np.dot(g, np.linalg.solve(h, g))
        return np.sqrt(dchisq)
        

    def _validate_bv(self, bv):
        """Helper method: converts 'bv' argument to a 1-d array of length B, and returns it."""

        bv = np.asarray(bv, dtype=float)       
        
        if (bv.ndim == 0) and (self.B == 1):
            return np.reshape(bv, (1,))
            
        if bv.shape != (self.B,):
            raise RuntimeError(f'Got {bv.shape=}, expected 1-d array of length {B=}')

        return bv

    
    ####################################################################################################
    #
    # "Fast" likelihood starts here.
    #
    # This code is completely unreadable, but there are unit tests which verify that the fast_*
    # functions are equivalent o their slow_* equivalents.


    def _init_fast_likelihood(self):
        S, V, K, D = self.S, self.V, self.K, self.D
        
        # xmu = shape (2,2,V,K), reshaped to (2,2*D).
        # Length-2 axes are fnl exponent and bv exponent.
        
        t = np.mean(self.surrs, axis=0)      # shape (2,2,D)
        self.xmu = np.reshape(t, (2,2*D))
        
        # xcov = shape (3,2,V,K,2,V,K), reshaped to (3,4*D*D)
        # Length-3 axis is fnl exponent {0,1,2}, and length-2 axes are bv exponents {0,1}.

        t = np.reshape(self.surrs, (S,4*D))      # shape (S, 4D)
        t = np.cov(t, rowvar=False)              # shape (4D,4D)
        t = np.reshape(t, (2,2*D,2,2*D))         # shape (2,2D,2,2D) where length-2 axes are fnl exponents
        t = np.array([ t[0,:,0,:], t[0,:,1,:]+t[1,:,0,:], t[1,:,1,:] ])   # shape (3,2D,2D)
        t = np.reshape(t, (3,4*D*D))         # shape (3,4D^2)
        self.xcov = np.copy(t)               # make contiguous


    def fast_mean_and_cov(self, fnl, bv, grad=False):
        """bv must be a 1-d array of length B, i.e. scalar is not allowed."""
        
        B, D, V, K = self.B, self.D, self.V, self.K
        f3 = np.array((1.0, fnl, fnl**2))

        bv = np.dot(bv, self.bias_matrix)  # shape (V,)
        bv20 = np.reshape(bv, (V,1))
        bv42 = np.reshape(bv, (1,1,V,1))
        bv50 = np.reshape(bv, (V,1,1,1,1))

        # Reminder: xmu = shape (2,2,V,K), reshaped to (2,2*D).
        mu0 = np.dot(f3[:2], self.xmu)       # shape (2*D)
        mu0 = np.reshape(mu0, (2,V,K))       # shape (2,V,K)
        mu = mu0[0,:,:] + bv20 * mu0[1,:,:]  # shape (V,K)
        mu = np.reshape(mu, (D,))

        # Reminder: xcov = shape (3,2,V,K,2,V,K), reshaped to (3,4*D*D)
        cov0 = np.dot(f3, self.xcov)                         # shape (4*D*D)
        cov0 = np.reshape(cov0, (2,V,K,2,V,K))               # shape (2,V,K,2,V,K)
        cov1 = cov0[0,:,:,:,:,:] + bv50 * cov0[1,:,:,:,:,:]  # shape (V,K,2,V,K)
        cov = cov1[:,:,0,:,:] + bv42 * cov1[:,:,1,:,:]       # shape (V,K,V,K)
        cov = np.reshape(cov, (D,D))

        if not grad:
            return mu, cov

        dmu_dfnl = np.reshape(self.xmu[1,:], (2,V,K))        # shape (2,V,K)
        dmu_dfnl = dmu_dfnl[0,:,:] + bv20 * dmu_dfnl[1,:,:]  # shape (V,K)
        dmu_dbv = mu0[1,:,:]                                 # shape (V,K)

        f2 = np.array((1.0, 2*fnl))
        dcov_dfnl = np.dot(f2, self.xcov[1:])                               # shape (4*D*D)
        dcov_dfnl = np.reshape(dcov_dfnl, (2,V,K,2,V,K))                    # shape (2,V,K,2,V,K)
        dcov_dfnl = dcov_dfnl[0,:,:,:,:,:] + bv50 * dcov_dfnl[1,:,:,:,:,:]  # shape (V,K,2,V,K)
        dcov_dfnl = dcov_dfnl[:,:,0,:,:] + bv42 * dcov_dfnl[:,:,1,:,:]      # shape (V,K,V,K)
        dcov_dbv0 = cov0[1,:,:,:,:,:]                                       # shape (V,K,2,V,K)
        dcov_dbv0 = dcov_dbv0[:,:,0,:,:] + bv42 * dcov_dbv0[:,:,1,:,:]      # shape (V,K,V,K)
        dcov_dbv1 = cov1[:,:,1,:,:]                                         # shape (V,K,V,K)
        
        mu_grad = np.zeros((B+1,V,K))     
        mu_grad[0,:,:] = dmu_dfnl
        mu_grad[1:,:,:] = self.bias_matrix.reshape((B,V,1)) * dmu_dbv.reshape((1,V,K))
        mu_grad = np.reshape(mu_grad, (B+1,D))
        
        cov_grad = np.zeros((B+1,V,K,V,K))
        cov_grad[0,:,:,:,:] = dcov_dfnl
        cov_grad[1:,:,:,:,:] = self.bias_matrix.reshape((B,V,1,1,1)) * dcov_dbv0.reshape((1,V,K,V,K))
        cov_grad[1:,:,:,:,:] += self.bias_matrix.reshape((B,1,1,V,1)) * dcov_dbv1.reshape((1,V,K,V,K))
        cov_grad = np.reshape(cov_grad, (B+1,D,D))
        
        return mu, cov, mu_grad, cov_grad


    def fast_log_likelihood(self, fnl, bv):
        if self.jeffreys_prior:
            # Need gradients
            mean, cov, grad_mean, grad_cov = self.fast_mean_and_cov(fnl, bv, grad=True)
        else:
            # No gradients needed
            mean, cov = self.fast_mean_and_cov(fnl, bv, grad=False)
        
        x = self.data_vector - mean
        l = np.linalg.cholesky(cov)
        linv_x = scipy.linalg.solve_triangular(l, x, lower=True)

        # log L = -(1/2) log(det C) - (1/2) x^T C^{-1} x
        logL = -0.5 * np.dot(linv_x, linv_x)
        logL -= np.sum(np.log(l.diagonal()))

        if self.jeffreys_prior:
            B, D = (self.B, self.D)
            linv_dmu = scipy.linalg.solve_triangular(l, grad_mean.T, lower=True)  # shape (D,2)
            f = np.dot(linv_dmu.T, linv_dmu)   # first term in 2-by-2 Fisher matrix

            # Second term in 2-by-2 Fisher matrix
            # F_{ij} = (1/2) Tr(C^{-1} dC_i C^{-1} dC_j)
            #        = (1/2) Tr(S_i S_j)  where S_i = L^{-1} dC_i L^{-T}
            
            t = grad_cov.reshape(((B+1)*D, D))
            u = scipy.linalg.solve_triangular(l, t.T, lower=True)  # shape (D, (B+1)*D)
            u = u.reshape((D*(B+1), D))
            v = scipy.linalg.solve_triangular(l, u.T, lower=True)  # shape (D,D*(B+1))
            v = v.reshape((D*D, B+1))
            f += 0.5 * np.dot(v.T, v)

            # Jeffreys prior is equivalent to including sqrt(det(F)) in the likelihood.
            sign, logabsdet_F = np.linalg.slogdet(f)
            logL += 0.5 * logabsdet_F
            assert sign == 1

        return logL

    
    ############################################  Testing  #############################################
    

    def test_fast_mean_and_cov(self):
        """Test fast_mean_and_cov(), by checking that it agrees with slow_mean_and_cov() at 10 random points."""

        for _ in range(10):
            fnl = np.random.uniform(-50, 50)
            bv = np.random.uniform(0, 1, size=(self.B,))
            
            slow_mean, slow_cov = self.slow_mean_and_cov(fnl, bv)
            slow_mean_grad, slow_cov_grad = self.slow_mean_and_cov_gradients(fnl, bv)
            fast_mean, fast_cov, fast_mean_grad, fast_cov_grad = self.fast_mean_and_cov(fnl, bv, grad=True)

            assert np.all(np.abs(slow_mean - fast_mean) < 1.0e-10)
            assert np.all(np.abs(slow_cov - fast_cov) < 1.0e-10)
            assert np.all(np.abs(slow_mean_grad - fast_mean_grad) < 1.0e-10)
            assert np.all(np.abs(slow_cov_grad - fast_cov_grad) < 1.0e-10)
    
    
    def test_fast_likelihood(self):
        """Test fast_log_likelihood(), by checking that it agrees with slow_log_likelihood() at 10 random points."""
        
        for _ in range(10):
            fnl = np.random.uniform(-50, 50)
            bv = np.random.uniform(0, 1, size=(self.B,))
            logL_slow = self.slow_log_likelihood(fnl, bv)
            logL_fast = self.fast_log_likelihood(fnl, bv)
            assert np.abs(logL_slow - logL_fast) < 1.0e-10
        
    
    @staticmethod
    def make_random():
        """Construct and return a PgvLikelihood with random (data, surrs, bias_matrix).
        
        Useful for standalone testing of 'class PgvLikelihood', in order to construct
        an "interesting" PgvLikelihood, in a situation where KSZ pipeline outputs are
        not available."""

        B = np.random.randint(1, 4)       # number of bias parameters in the MCMC
        V = np.random.randint(B, B+3)     # number of velocity reconstructions
        K = np.random.randint(5, 15)      # number of k-bins
        S = np.random.randint(100, 200)   # number of surrogate sims
        jeffreys_prior = (np.random.uniform() < 0.5)   # boolean

        data = np.random.normal(size=(V,K))
        surrs = np.random.normal(size=(S,2,2,V,K))

        # Randomly generate the bias_matrix (shape (B,V), where B <= V)
        # This is not so straightforward, since we want to avoid small SVD eigenvalues
        # for numerical stability.

        rot1 = utils.random_rotation_matrix(B)
        rot2 = utils.random_rotation_matrix(V)
        svds = np.random.uniform(1.0, 2.0, size=B)
        bias_matrix = np.dot(rot1, svds.reshape((B,1)) * rot2[:B,:])

        return PgvLikelihood(data, surrs, bias_matrix, jeffreys_prior)

        
    @staticmethod
    def run_tests():
        """Runs standalone tests of 'class PgvLikelihood'.
        (Where "standalone" means that no KSZ pipeline outputs are needed.)"""
        
        for _ in range(20):
            lik = PgvLikelihood.make_random()
            lik.test_fast_mean_and_cov()
            lik.test_fast_likelihood()

        print('PgvLikelihod.run_tests(): pass')



from . import core
from . import utils

from .Box import Box
from .Cosmology import Cosmology
from .Catalog import Catalog

import numpy as np
from collections.abc import Iterable


class CatalogGridder:
    def __init__(self, box, cosmo, randcat, rweights, nfootprints, zcol_name='z', save_rmaps=False, save_fmaps=False, kernel='cubic'):
        r"""CatalogGridder: operates on fields $f(x)$ which are weighted sums over catalogs, and outputs Fourier-space maps.

        This is probably not the class you're looking for -- you probably want :class:`~kszx.KszPSE` instead!
        CatalogGridder is a "mid-level" class which is higher level than :func:`~kszx.estimate_power_spectrum()`,
        but is intended to b a building block for higher-level classes (:class:`~kszx.KszPSE` and perhaps
        eventually a GalaxyPSE class).

        The purpose of the CatalogGridder is to grid fields to Fourier-space maps, keeping track of
        normalizations, so that estimated power spectra have normalizations which are approximately
        correct (see below). In this docstring, we give an informal description -- for a formal
        specification, see "CatalogGridder :ref:`catalog_gridder_details`" in the sphinx docs.

        We distinguish between two types of fields which are weighted sums over catalogs:

          1. A "density" field is a weighted sum of tracers, with randoms subtracted:

             $$\delta(x) \propto \sum_i W_i \delta^3(x-x_i) - (\mbox{randoms})$$
 
          2. A "sampled" field is obtained from an underlying continuous field $F(x)$,
             by sampling at discrete points $x_i$ with sampling weights $w_i$:

             $$f(x) = \sum_i c_i \delta^3(x-x_i) \hspace{1cm} \mbox{where } c_i = w_i F(x_i) + (\mbox{noise})$$
        
        For a density field, we're interested in clustering statistics of the tracers themselves.
        For a sampled field, we're interested in clustering statsitics of the underlying continuous 
        field $F(x)$. A galaxy field $\delta_g(x)$ is a good example of a density field, and the
        kSZ velocity reconstruction $\hat v_r(x)$ is a good example of a sampled field. 

        For sampled fields, the $(c_i, w_i, F(x))$ notation deserves more explanation.
        In the following bullet points, we consider a concrete example: kSZ velocity reconstruction.

          - Recall from the overleaf that kSZ velocity reconstruction $\hat v_r$ is defined by
        
            $$\hat v_r(x) = \sum_i W_i^S \tilde T_{\rm CMB}(\theta_i) \delta^3(x-x_i)$$
        
            where $W_i^S$ is a per-galaxy weighting (e.g. FKP), and $\tilde T(\theta)$ is the 
            filtered CMB. 

          - We assume that, on large scales, the relation between $\langle \hat v_r \rangle$ 
            and $v_r^{\rm true}$ is:

            $$\langle v_r(x) \rangle = \sum_i W_i^S b_i^v v_r^{\rm true}(x_i) \delta^3(x-x_i)$$
        
          - In this example, the "underlying continuous field" $F(x)$ is the radial velocity 
            field $v_r(x)$. Comparing the last three equations, the coefficients $c_i$ and
            sampling weights $w_i$ are given by:
        
            $$c_i = W_i^S \tilde T_{CMB}(\theta_i)
            \hspace{1.5cm}
            w_i = W_i^S b_i^v$$

            We emphasize that the "sampling weight" $w_i$ is not equal to the galaxy weight $W_i^S$.
            Instead, we have $w_i = W_i^S b_i^v$, which is not initially obvious!
        
        The CatalogGridder is used as follows:

          - When a CatalogGridder is constructed, the caller specifies $N_{foot}$ "footprints". 
            Each footprint is specified by a random catalog (``randcat``) with weights (``rweights``).

          - For a density field, the constructor ``rweights`` should reflect the weighting
            that will later be applied to the galaxies (i.e. the ``gweights`` argument to 
            :meth:`~kszx.CatalogGridder.grid_density_field`). For example, if galaxies will
            be FKP-weighted, then the constructor ``rweights`` should be FKP weights for the 
            random catalog. Warning: even a small mismatch between the weighting of
            galaxies/randoms can produce a large contribution to the power spectrum!

          - For a sampled field, the constructor ``rweights`` should reflect the sampling weights $w_i$
            (defined above) that will later be applied to the galaxies. For example, in the case of
            the kSZ velocity reconstruction $\hat v_r$, we have $w_i = W_i^S b_i^v$ (not obvious,
            see above!!). Therefore, the constructor ``rweights`` should be $(W_i^S b_i^v)$,
            evaluated on the random catalog.

            (Note that for the specific case of kSZ velocity reconstruction, there is a class 
            :class:`~kszx.KszPSE` which handles these details automatically.)

          - The overall normalization of ``rweights`` is unimportant (in both the "density field"
            and "sampled field" cases).
        
          - After constructing a CatalogGridder, fields are "gridded" to Fourier space maps by
            calling :meth:`~kszx.CatalogGridder.grid_density_field` for a density field, or 
            :meth:`~kszx.CatalogGridder.grid_sampled_field` for a sampled field.

            When fields are gridded, a ``footprint_index`` is specified, to provide the appropriate
            normalization. In the case of a density field, random subtraction is also performed.
        
            WARNING: Fourier-space maps returned by these methods have an unusual normalization,
            which is chosen to give sensible power spectrum estimates (see next bullet point), but
            may not be what you expect! For more info, see "CatalogGridder :ref:`catalog_gridder_details`" 
            in the sphinx docs.

          - After making Fourier-space maps, you compute "unnormalized" power spectra by calling
            the low-level function :func:`~kszx.estimate_power_spectrum`, and then normalize
            by multiplying by ``catalog_gridder.ps_normalization[i,j]``, where $i$ and $j$ are
            the footprint indices.

            The ``ps_normalization`` member is computed in the constructor, and is a matrix $N_{ij}$
            which can be used to normalize the power spectrum for gridded fields on footprints $i,j$.
            This normalization is an ansatz which should be approximately correct, but may break down
            on large scales. (Eventually, we'll implement rigorous window function deconvolution.)

        When constructor arguments are described in the bullet points below, we sometimes use pluralized
        notation. For example, the ``randcat`` arg has type kszx.Catalog(s). This means that the ``randcat``
        arg can either be a length-``nfootprints`` list of Catalog objects, or a single Catalog object.
        (Specifying a single Catalog is equivalent to repeating the Catalog ``nfootprints`` times.)
        
        Constructor arguments:

           - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
             See :class:`~kszx.Box` for more info.

           - ``cosmo`` (kszx.Cosmology): see :class:`~kszx.Cosmology`.
             Only used to convert redshifts to distances.
        
           - ``randcat`` (kszx.Catalog(s)): see next item.

           - ``rweights`` (scalar(s) or 1-d arrays): as described above, the CatalogGridder
             contains $N_{foot}$ footprints. Each footprint is specified by a random catalog
             (``randcat``) with weights (``rweights``).

             The ``rweights`` for each footprint can either be a 1-d array of length ``randcat.size``,
             or a scalar (if every object has the same weight). See above for more discussion of
             how to choose ``rweights``.
        
           - ``nfootprints`` (int): Number of footprints $N$.
        
           - ``zcol_name`` (string(s)): Each ``randcat`` must contain columns ``'ra_deg'``,
             ``'dec_deg'``, and a redshift column which is ``'z'`` by default, but this can
             be overridden by specifying the ``zcol_name`` argument.

             For example, in a photometric randcat, you might have two columns ``'ztrue'``
             and ``'zobs'``. In this case, you probably want ``zcol_name = 'zobs'``, in
             order to grid catalogs at observed redshifts, not true (unobserved) redshifts.
        
           - ``save_rmaps`` (boolean(s)): If ``True`` for a footprint, then a real-space gridded
             footprint will be saved. **This is needed in order to call** :meth:`grid_density_field`
             **on the footprint.** It consumes significant memory, so you may want to set ``save_rmaps``
             to True only on footprints where ``grid_density_field()`` will be called.

             (If you call ``grid_density_field()`` but forgot to set ``save_rmaps=True``, then
             you'll get a verbose error message.)
        
           - ``save_fmaps`` (boolean(s)): If ``True`` for a footprint, then a Fourier-space
             gridded footprint will be saved. This is only useful for debugging.
        
           - ``kernel`` (string): either ``'cic'`` or ``'cubic'`` (more options will be defined later).

        Class members:

           - ``ps_normalization`` (matrix): An $N_{\rm foot}$-by-$N_{\rm foot}$ matrix $N_{ij}$.

             Multiplying by $N_{ij}$ is the final step in computing normalized power spectrum
             estimates. (After computing Fourier-space maps with :meth:`grid_density_field()`
             and/or :meth:`grid_sampled_field()`, then computing unnormalized power spectra
             the low-level function :func:`~kszx.estimate_power_spectrum()`.)
        """
        
        assert isinstance(box, Box)
        assert isinstance(cosmo, Cosmology)
        assert kernel in ['cic','cubic']  # more kernels coming soon
        assert int(nfootprints) == nfootprints
        assert nfootprints >= 1
        assert nfootprints <= 50

        self.box = box
        self.cosmo = cosmo
        self.nfootprints = int(nfootprints)
        self.kernel = kernel
        
        # "Vector" args:
        #   randcats     length-nfootprints list of Catalogs
        #   zcol_names   length-nfootprints list of strings
        #   save_rmaps   length-nfootprints list of booleans
        #   save_fmaps   length-nfootprints list of booleans
        #   rweights     length-nfootprints list of (0-d or 1-d numpy array)

        self.randcats = self._parse_vector_arg(randcat, 'randcat', Catalog)
        self.zcol_names = self._parse_vector_arg(zcol_name, 'zcol_name', str)
        self.save_rmaps = self._parse_vector_arg(save_rmaps, 'save_rmaps', bool)
        self.save_fmaps = self._parse_vector_arg(save_fmaps, 'save_fmaps', bool)
        self.rweights = self._parse_rweights(rweights)

        assert all(rcat.size >= 1000 for rcat in self.randcats)
        
        self.rmaps = [ None for _ in range(self.nfootprints) ]  # real-space maps
        self.fmaps = [ None for _ in range(self.nfootprints) ]  # Fourier space maps
        self.rwsum = np.zeros(self.nfootprints)

        # FIXME this loop can be optimized, to avoid repeating computations in cases
        # where different fields use the same (randcat, rweights, zcol_name).
        
        for i in range(self.nfootprints):
            rc = self.randcats[i]
            rw = self.rweights[i]
            points = rc.get_xyz(cosmo, zcol_name = self.zcol_names[i])
            
            assert (rw.ndim == 0) or (rw.shape == (rc.size,))
            rw = rw if (rw.ndim > 0) else np.full(rc.size, rw)
                
            self.rwsum[i] = np.sum(rw)
            self.rmaps[i] = core.grid_points(box, points, rw, kernel=kernel)
            self.fmaps[i] = core.fft_r2c(box, self.rmaps[i])

            assert self.rwsum[i] != 0
            core.apply_kernel_compensation(box, self.fmaps[i], kernel)

            if not self.save_rmaps[i]:
                self.rmaps[i] = None

        # FIXME revisit the issue of choosing K!
        K = 0.6 * box.knyq
        self.A = self.compute_A(K)
        assert self.A.shape == (self.nfootprints, self.nfootprints)
        assert np.all(self.A.diagonal() > 0)

        # Clear fmaps after calling self.compute_A().
        for i in range(self.nfootprints):
            if not self.save_fmaps[i]:
                self.fmaps[i] = None
            
        self.ps_normalization = np.zeros((self.nfootprints, self.nfootprints))
        
        for i in range(self.nfootprints):
            for j in range(i+1):
                r = self.A[i,j] / np.sqrt(self.A[i,i] * self.A[j,j])
                if np.abs(r) < 0.03:
                    print(f'CatalogGridder: fields {(j,i)} have non-overlapping footprints, cross power will not be estimated')
                else:
                    self.ps_normalization[i,j] = self.ps_normalization[j,i] = 1.0 / self.A[i,j]


    def grid_density_field(self, gcat, gweights, footprint_index, zcol_name='z'):
        r"""Grids a density field, subtracts randoms (with footprint_index), and returns a Fourier-space map.

        The density field is specified by points $x_i$ (via the ``gcat`` and ``zcol_name`` args),
        and weights $W_i$ (via the ``gweights`` arg), and is defined by:

        $$\delta(x) \propto \sum_i W_i \delta^3(x-x_i) - \big( \mbox{randoms} \big)$$

        where we have swept the normalization under the rug by using the $\propto$ symbol.
        The grid_density_field() method will assign a normalization which is appropriate if you're
        interested in clustering statistics of the *overdensity* field $\delta = \rho/\bar\rho$.
        For a precise specification of the normalization, see 
        "CatalogGridder :ref:`catalog_gridder_details`" in the sphinx docs.
        
        Function arguments:

          - ``gcat`` (:class:`Catalog`): Defines spatial locations $x_i$.

          - ``gweights`` (scalar, or 1-d numpy array of length ``gcat.size``):
            Defines weights $W_i$.
        
          - ``footprint_index`` (integer): An integer $0 \le i < N_{\rm footprints}$.
            Specifies which "footprint" (see constructor docstring) is used to subtract
            randoms, and for normalization.

          - ``zcol_name``` (string):  The catalog ``gcat`` must contain columns ``'ra_deg'``,
            ``'dec_deg'``, and a redshift column which is ``'z'`` by default, but this can
            be overridden by specifying the ``zcol_name`` argument.

            For example, in a photometric catalog, you might have two columns ``'ztrue'``
            and ``'zobs'``. In this case, you probably want ``zcol_name = 'zobs'``, in
            order to put tracers at their observed redshifts, not true (unobserved) redshifts.
        """
        
        assert isinstance(gcat, Catalog)
        assert 0 <= footprint_index < self.nfootprints

        if not self.save_rmaps[footprint_index]:
            raise RuntimeError(f"CatalogGridder.grid_density_field() was called with {footprint_index=},"
                               + f" but save_rmaps=False for this footprint. You probably want to set"
                               + f" save_rmaps=True in the constructor, but only for footprint indices"
                               + f" where grid_density_field() will be called. See docstring for more info.")
        
        rwsum = self.rwsum[footprint_index]
        points = gcat.get_xyz(self.cosmo, zcol_name=zcol_name)
        gweights = self._parse_coeffs(gcat, gweights, 'CatalogGridder.grid_density_field()', 'gweights')

        wsum = np.sum(gweights)
        assert wsum != 0.0

        rmap = core.grid_points(self.box, points, (rwsum/wsum) * gweights, kernel=self.kernel)
        rmap -= self.rmaps[footprint_index]   # subtract randoms

        fmap = core.fft_r2c(self.box, rmap)
        core.apply_kernel_compensation(self.box, fmap, self.kernel)
        return fmap


    def grid_sampled_field(self, gcat, coeffs, wsum, footprint_index, spin=0, zcol_name='z'):
        r"""Grids a sampled field $f(x) = \sum_i c_i \delta^3(x-x_i)$, and returns a Fourier-space map.

        A "sampled" field is obtained from an underlying continuous field $F(x)$, by sampling
        at discrete points $x_i$ with weights $w_i$:

        $$f(x) = \sum_i c_i \delta^3(x-x_i) \hspace{1cm} \mbox{where } c_i = w_i F(x_i) + (\mbox{noise})$$

        For example, kSZ velocity reconstruction (summarizing discussion from the class 
        :class:`~kszx.CatalogGridder` docstring):

        $$\begin{align}
        \hat v_r(x) &= \sum_i W_i^S \tilde T_{\rm CMB}(\theta_i) \delta^3(x-x_i) \\
        c_i &= W_i^S \tilde T_{CMB}(\theta_i) \\
        w_i &= W_i^S b_i^v
        \end{align}$$

        (Note that for the specific case of kSZ velocity reconstruction, there is a higher-level class
        :class:`~kszx.KszPSE`, and you shouldn't need to use :class:`~kszx.CatlogGridder` directly.)

        The grid_sampled_field() method will assign a normalization which is appropriate if you're
        interested in clustering statistics of the *underlying continuous* field $F(x)$.
        For a precise specification of the normalization, see 
        "CatalogGridder :ref:`catalog_gridder_details`" in the sphinx docs.

        Function arguments:

          - ``gcat`` (:class:`Catalog`): Defines spatial locations $x_i$.

          - ``coeffs`` (scalar, or 1-d numpy array of length ``gcat.size``):
            Coefficients $c_i$ appearing in the sampled field $f(x) = \sum_i c_i \delta^3(x-x_i)$.

          - ``wsum`` (scalar): The sum of the sampling weights $\sum_i w_i$.
            This is needed to assign a normalization to the output Fourier-space map.
            (Note that we don't need the individual sampling weights $w_i$, just their sum.)

          - ``footprint_index`` (integer): An integer $0 \le i < N_{\rm footprints}$.
            Specifies which "footprint" (see constructor docstring) is used to compute
            the normalization.

          - ``spin`` (integer): The "spin" of the FFT called at the end of grid_sampled_field()
            to go from real space to Fourier space. You should set ``spin=0`` for a scalar field
            (e.g. surrogate galaxy field $S_g(x)$), and ``spin=1`` for a radial velocity (e.g.
            kSZ velocity reconstruction $\hat v_r(x)$. For the precise definition of ``spin``,
            see FFT :ref:`fft_conventions` in the sphinx docs.

          - ``zcol_name`` (string):  The catalog ``gcat`` must contain columns ``'ra_deg'``,
            ``'dec_deg'``, and a redshift column which is ``'z'`` by default, but this can
            be overridden by specifying the ``zcol_name`` argument.

            For example, in a photometric catalog, you might have two columns ``'ztrue'``
            and ``'zobs'``. In this case, you probably want ``zcol_name = 'zobs'``, in
            order to put tracers at their observed redshifts, not true (unobserved) redshifts.
        """
        
        assert isinstance(gcat, Catalog)
        assert 0 <= footprint_index < self.nfootprints
        assert wsum != 0.0

        rwsum = self.rwsum[footprint_index]
        points = gcat.get_xyz(self.cosmo, zcol_name=zcol_name)
        coeffs = self._parse_coeffs(gcat, coeffs, 'CatalogGridder.grid_sampled_field()', 'coeffs')

        fmap = core.grid_points(self.box, points, (rwsum/wsum) * coeffs, kernel=self.kernel, fft=True, spin=spin)
        core.apply_kernel_compensation(self.box, fmap, self.kernel)
        return fmap
        

    def compute_A(self, K):
        r"""Computes the $A$-matrix (an intermediate quantity used to compute PS normalization). Intended for debugging!

        You shouldn't need to call this function, unless you're interested in the details of the power spectrum
        normalization! For a definition of what it computes, see "CatalogGridder :ref:`catalog_gridder_details`" 
        in the sphinx docs, where the $A$-matrix is denoted $A_{RR'}$.

        Returns a shape (nfootprints, nfootprints) array. In order to call ``compute_A()``, you need to call 
        the ``CatalogGridder`` constructor with ``save_fmaps=True``."""

        if any(x is None for x in self.fmaps):
            raise RuntimeError('CatalogGridder: in order to call compute_A(), you should call'
                               + ' the CatalogGridder constructor with save_fmaps=True')
            
        edges = np.array([ 0, K / 2**(1./self.box.ndim), K ])

        # use_dc=True is important here!!
        pk, counts = core.estimate_power_spectrum(self.box, self.fmaps, edges, use_dc=True, return_counts=True)
        assert pk.shape == (self.nfootprints, self.nfootprints, 2)
        assert counts.shape == (2,)

        # Note factor (1/V_box), which normalizes sum_k -> int d^3k/(2pi)^3
        return (pk[:,:,0] - pk[:,:,1]) * counts[0] / self.box.box_volume


    ####################################################################################################
    

    def _parse_vector_arg(self, arg, arg_name, _type):
        if isinstance(arg, _type):
            return [ arg ] * self.nfootprints

        try:
            n = len(arg)
        except:
            raise RuntimeError(f'CatalogGridder: expected {arg_name} to be either {type}, or list of {type}')

        if n != self.nfootprints:
            raise RuntimeError(f'CatalogGridder: expected {arg_name} to have length nfootprints={self.nfootprints}, got length {n}')

        arg = list(arg)
        assert len(arg) == n

        for i in range(n):
            if not isinstance(arg[i], _type):
                raise RuntimeError(f'CatalogGridder: expected {arg_name} to be either {type}, or list of {type}')                

        return arg


    def _parse_rweights(self, rweights):
        if self._is_scalar(rweights):
            rweights = np.array(float(rweights))   # 0-d numpy array
            return [ rweights ] * self.nfootprints

        if isinstance(rweights,np.ndarray) and all(rweights.shape == (rcat.size,) for rcat in self.randcats):
            rweights = utils.asarray(rweights, 'CatalogGridder', 'rweights', dtype=float)
            return [ rweights ] * self.nfootprints

        try:
            n = len(rweights)
        except:
            raise RuntimeError("CatalogGridder: couldn't parse 'rweights' (neither iterable nor scalar)")

        if n != self.nfootprints:
            raise RuntimeError(f"CatalogGridder: couldn't parse 'rweights' (got length {n}, expected"
                               + f" length nfootprints={self.nfootprints})")
        
        rweights = [ utils.asarray(w, 'CatalogGridder', 'rweights', dtype=float) for w in rweights ]
        assert len(rweights) == self.nfootprints

        for i in range(self.nfootprints):
            if (rweights[i].ndim > 0) and (rweights[i].shape != (self.randcats[i].size,)):
                raise RuntimeError(f"CatalogGridder: expected rweights[{i}] to be a scalar or length"
                                   + f" randcats[{i}].size={self.randcats[i].size}, got shape" 
                                   + f" {rweights[i].shape}")

        return rweights
    
            
    @staticmethod
    def _is_scalar(f):
        try:
            f = float(f)
            return True
        except:
            pass
        return False


    def _parse_coeffs(self, gcat, coeffs, func_name, arg_name):
        """Helper for grid_density_field() and grid_sampled_field()."""
        
        coeffs = utils.asarray(coeffs, func_name, arg_name, dtype=float)
        
        if coeffs.shape == (gcat.size,):
            return coeffs
        if coeffs.ndim == 0:
            return np.full(gcat.size, coeffs)
        
        raise RuntimeError(f"{func_name}: expected {arg_name} to have length {gcat.size}"
                           + f" (=gcat.size), got shape {coeffs.shape}")



from . import cpp_kernels

from .core import \
    fft_r2c, \
    fft_c2r, \
    interpolate_points, \
    grid_points, \
    apply_kernel_compensation, \
    multiply_rfunc, \
    multiply_kfunc, \
    multiply_r_component, \
    apply_partial_derivative, \
    simulate_white_noise, \
    simulate_gaussian_field, \
    estimate_power_spectrum, \
    kbin_average, \
    fkp_from_ivar_2d, \
    ivar_combine, \
    estimate_cl

from . import utils
from . import io_utils
from . import healpix_utils
from . import pixell_utils

from .Box import Box
from .BoundingBox import BoundingBox
from .Catalog import Catalog
from .Cosmology import Cosmology, CosmologicalParams

from .CmbClFitter import CmbClFitter
from .CatalogGridder import CatalogGridder
from .KszPSE import KszPSE

from . import act
from . import desi
from . import desils_lrg
from . import desils_main
from . import ksz_desils
from . import planck
from . import plot
from . import sdss
from . import tests



from . import utils
from . import pixell_utils
from . import plot

from .Cosmology import Cosmology

import pixell
import numpy as np
import scipy.optimize
import matplotlib.pyplot as plt


class CmbClFitter:
    def __init__(self, cosmo, cmb_map, weight_map, bl, lmin, lmax, ivar=None, iterative=False, fit_cmb_amplitude=False, fit_alpha=True, fit_lred=True, beamed_red_noise=True, uniform_lweight=False, alpha_fid=-3, lred_fid=1500, lpad=1000):
        r"""Given a CMB map, fit a power spectrum of the form $C_l=$ (lensed Cls) + (power-law red noise) + (white noise).
        
        Based on utils.get_single_frequency_alms() in Alex + Mat's kSZquest software:
            https://github.com/alexlague/kSZquest/blob/main/utils.py

        An example notebook where CmbClFitter is used:
            https://github.com/kmsmith137/kszx_notebooks/blob/main/05_sdss_pipeline/03_prepare_cmb_weighting.ipynb

        With default constructor arguments, the CmbClFitter fits the power spectrum of the
        specified ``cmb_map`` (with pixel weighting specified by ``weight_map``, and for 
        $l_{\rm min} \le l \le l_{\rm max}$) to a model of the form:

        $$\begin{align}
        C_l &= b_l^2 C_l^{\rm CAMB} + A_{\rm red} b_l^2 \max\left( \frac{l}{l_{\rm red}}, 1 \right)^\alpha + A_{\rm white}
        \end{align}$$
        with parameters $(A_{\rm red}, A_{\rm white}, \alpha, l_{\rm red})$. Here, $C_l^{CAMB}$ is the 
        **lensed** CMB power spectrum, and does not include kSZ or foregrounds.

        I experimented with a lot of flags that turned out not to really matter. The default
        constructor arguments above (with ``lmin=1500`` and ``lmax=8000``) are what I've been
        using for ACT. 

        If you want to reproduce Alex + Mat's fits from kSZquest, use the following arguments 
        (only args which differ from default values are shown):

           - lmin = 1000
           - lmax = 8000
           - fit_cmb_amplitude = True
           - fit_alpha = False
           - fit_lred = False
           - beamed_red_noise = False
           - uniform_lweight = True

        **Note:** CmbClFitter currently uses pixell maps (for ``cmb_map``, ``weight_map``, ``ivar``), but could
        easily be modified to allow healpix maps -- let me know if this would be useful.
            
        The output of the fitting procedure is contained in the following class members, which are initialized
        by the constructor::

            # Scalar params
            self.alpha           # spectral index of red noise (see above)
            self.red_ampl        # coefficient A_red of red noise contribution to C_l (see above)
            self.write_ampl      # coefficient A_white of white noise contribution to C_l (see above)
            self.cmb_ampl        # coefficient A_cmb of CMB contribution to C_l (see 'fit_cmb_amplitude' below)
            self.l_knee          # value of l where white/red noise are equal
            self.uK_arcmin       # equivalent to A_white, just reparameterized as noise RMS
            self.ivar_uK_arcmin  # equivalent white noise level of ivar map (if ivar is specified)

            # Model power spectra
            self.cl_beamed_cmb    # beam-convolved, noise-free
            self.cl_red_noise     # power-law in l (possibly beam-convolved, see 'beamed_red_noise' below)
            self.cl_white_noise   # constant in l
            self.cl_tot           # sum of previous 3 contributions

            # Weighted or "pseudo" power spectra. These are used by CmbClFitter.plot(), but are probably
            # not useful otherwise. For now, I'm converting model Cls to pseudo Cls by running on simulation.
            # Some day I'll improve this, by implementing the pseudo-Cl transfer matrix.
            
            self.pcl_beamed_cmb   # beam-convolved, noise-free, from one sim
            self.pcl_red_noise    # power spectrum of red noise, from one sim
            self.pcl_white_noise  # constant in l
            self.pcl_tot          # sum of previous 3 contributions
            self.pcl_data         # empirical spectrum of cmb map

        Constructor arguments:

          - ``cosmo`` (:class:`~kszx.Cosmology`): only needed for CAMB lensed $C_l$s.

          - ``cmb_map`` (pixell.enmap.ndmap): CMB temperature map with units $\mu$K
            (e.g. from calling :func:`~kszx.act.read_cmb()`).

          - ``weight_map`` (pixell.enmap.ndmap): applied to all maps before computing
            alms or cls. Recommend including a foreground mask, multiplied by some 
            downweighting of noisy regions (e.g. hard cutoff, ivar weighting, or FKP
            weighting).

          - ``bl`` (1-d numpy array): beam transfer function (e.g. from calling 
            :func:`~kszx.act.read_ivar()`).

          - ``lmin``, ``lmax`` (integers): range of $l$-values over which fit is performed.
            For ACT I've been using ``lmin=1500`` and ``lmax=8000`` (note that the CmbClFitter
            is intended to fit high $l$-values).

          - ``ivar`` (pixell.enmap.ndmap, optional): inverse variance map (e.g. from calling
            :func:`~kszx.act.read_ivar()`). This is not used in the fitting procedure! If ivar
            is specified, then the constructor will initialize ``self.ivar_uK_arcmin``, which
            can be compared to ``self.uK_arcmin``, to get a sense for how well the fitted white
            noise level $A_{\rm white}$ compares to the ivar map.

          - ``iterative`` (boolean, default False): If False, then we assume that pseudo-Cls and
            true Cls are related by a factor $W_2 = \int d^2\theta W(\theta)^2/4\pi$, where $W(\theta)$
            is the ``weight_map``. If True, then we try to improve this approximation using a Monte
            Carlo based, iterative approach. This turned out to make almost no difference, so I don't
            recommend that you use ``iterative=True``!

          - ``fit_cmb_amplitude`` (boolean, default False): If True, then the amplitude of the
            CMB power spectrum is a free (fitted) parameter, i.e. the CMB term is 
            $(A_{\rm cmb} b_l^2 C_l^{\rm CAMB})$ rather than $(b_l^2 C_l^{\rm CAMB})$.

          - ``fit_alpha`` (boolean, default True): If True, then the spectral index $\alpha$
            of the red noise is a free (fitted) parameter. If False, then $\alpha$ is fixed
            to the value ``alpha_fid`` (another constructor argument, see below).

          - ``fit_lred`` (boolean, default True): The $l_{\rm red}$ parameter regulates the
            red noise at low $l$ (see equation near the beginning of this docstring).
            If ``fit_red=True``, then $l_{\rm red}$ is a free (fitted) parameter. If ``fit_red=False``,
            then $l_{\rm red}$ is fixed to the value ``lred_fid`` (another constructor argument,
            see below).

          - ``beamed_red_noise`` (boolean, default True): If True, then the red noise is a 
            beam-convolved power law $C_l = (A_{\rm red} b_l^2 \max(l/l_{\rm red},1)^\alpha)$.
            If False, then the red noise is a power law $C_l = (A_{\rm red} \max(l/l_{\rm red},1)^\alpha)$.

          - ``uniform_lweight`` (boolean, default False): determines how $\chi^2$ is defined 
            (the fit is implemented by minimizing $chi^2$).

               - if False, then $\chi^2 = \sum_{l=l_{min}}^{l_{max}} (\Delta C_l)^2$.

               - if True, then $\chi^2 = \sum_{l=l_{min}}^{l_{max}} (\Delta C_l)^2/(l C_l^2)$.

          - ``alpha_fid`` (float, default -3): Fiducial value of red noise spectral index $\alpha$.
            (If ``fit_alpha=True``, then the value of ``alpha_fid`` shouldn't matter -- we only use it
            to set initial parameter values for the $\chi^2$ minimization.)

          - ``lred`` (float, default 1500): Fiducial value of the $l_{\rm red}$ parameter, which
            regulates the red noise at low $l$ (see equation near the beginning of this docstring).
            (if ``fit_lred=True``, then the value of ``lred`` shouldn't matter -- we only use it to
            to set initial parameter values for the $\chi^2$ minimization.)
        
          - ``lpad`` (integer, default 1000): After doing the fit, we validate the fit by comparing
            the empirical pseudo-$C_l$s of the data to a simulation of the model. We simulate the
            CMB to a multipole $(l_{\rm max} + l_{\rm pad})$ which is larger than the maximum
            multipole $l_{\rm max}$ of the power spectrum estimation/fitting.
        """

        # Argument checking starts here.
        
        bl = utils.asarray(bl, 'CmbClFitter', 'bl', dtype=float)
        
        assert isinstance(cosmo, Cosmology)
        assert isinstance(cmb_map, pixell.enmap.ndmap)
        assert isinstance(weight_map, pixell.enmap.ndmap)
        assert cmb_map.ndim == 2
        assert weight_map.ndim == 2
        assert 2 <= lmin < lmax
        assert lmin <= lred_fid < lmax
        assert alpha_fid < 0
        assert bl.ndim == 1
        assert lpad > 0
        
        beam_lmax = len(bl)-1
        assert beam_lmax >= (lmax + lpad)
        
        if ivar is not None:
            assert isinstance(ivar, pixell.enmap.ndmap)
            assert ivar.ndim == 2
            
        if cosmo.lmax < (lmax + lpad):
            raise RuntimeError(f"CmbClFitter: {cosmo.lmax=} must be >= {(lmax+lpad)=}. To increase cosmo.lmax,"
                               + f" call the Cosmology constructor with the 'lmax' argument specified")
        
        # Argument checking ends here.
        
        self.lmin = lmin
        self.lmax = lmax
        self.lpad = lpad
        self.lred = lred_fid
        self.alpha = alpha_fid
        self.fit_lred = fit_lred
        self.fit_alpha = fit_alpha
        self.fit_cmb_amplitude = fit_cmb_amplitude
        self.beamed_red_noise = beamed_red_noise
        self.uniform_lweight = uniform_lweight
        self.iterative = iterative
        self.bl = bl[:(lmax+lpad+1)]
        
        # self._init_white_noise() initializes the following members:
        #
        #  self.w2               # approximate normalization used to compute pseudo-Cls
        #  self.ivar_uK_arcmin   # equivalent noise level of ivar map
        
        self._init_white_noise(weight_map, ivar)

        # _pcl_from_weighted_map() must come after _init_white_noise()
        self.pcl_data = self._pcl_from_weighted_map(cmb_map * weight_map)
        
        # self._do_fit() initializes or updates the following members:
        #
        #  self.alpha           # current value of alpha
        #  self.red_ampl        # current value of A in C_l = A (l/lmin)^alpha
        #  self.white_ampl      # current value of B in C_l = B
        #  self.cmb_ampl        # multiplier of C_l^{lensed}
        #  self.red_template    # pseudo-cls corresponding to (l/lmin)^alpha, from one sim
        #  self.cmb_template    # pseudo-cls corresponding to C_l^{lensed}, from one sim
        #  self.l_knee          # value of l where white/red noise are equal
        #  self.uK_arcmin       # reparameterization of self.white_ampl

        niter = 2 if iterative else 1
        for _ in range(niter):
            self._do_fit(cosmo, weight_map)
        
        # Done -- just need to initialize remaining members.
        
        self.cl_beamed_cmb = self.cmb_ampl * cosmo.cltt_len[:(lmax+1)] * bl[:(lmax+1)]**2
        self.cl_red_noise = self.red_ampl * self._cl_red_noise()
        self.cl_white_noise = np.full(lmax+1, self.white_ampl)
        self.cl_tot = self.cl_beamed_cmb + self.cl_red_noise + self.cl_white_noise

        self.pcl_beamed_cmb = self.cmb_ampl * self.cmb_template
        self.pcl_red_noise = self.red_ampl * self.red_template
        self.pcl_white_noise = np.full(lmax+1, self.white_ampl)
        self.pcl_tot = self.pcl_beamed_cmb + self.pcl_red_noise + self.pcl_white_noise
        # Reminder: self.pcl_data was initialized above.
        
         
    def make_plots(self, filename_prefix=None, suffix='pdf', s=30):
        """Make three diagnostic plots, either on-screen (filename_prefix=None) or saved to disk.
        
          - Plot 1: pseudo-Cls, data vs model (filename ``{filename_prefix}_pcl.{suffix}``).
          - Plot 2: pseudo-Cls, data/model ratio (filename ``{filename_prefix}_pcl_ratio.{suffix}``).
          - Plot 3: Model Cls (filename ``{filename_prefix}_cl.{suffix}``).
        """

        filename = (filename_prefix + '_pcl' + suffix) if (filename_prefix is not None) else None
        self._plot_smoothed(self.pcl_data, s, label='data', l2=True)
        self._plot_smoothed(self.pcl_tot, s, label='model (total)', l2=True)
        self._plot_smoothed(self.pcl_beamed_cmb, s, label='model (cmb)', l2=True)
        self._plot_smoothed(self.pcl_red_noise, s, label='model (red)', l2=True)
        self._plot_smoothed(self.pcl_white_noise, s, label='model (white)', l2=True)
        plt.title(r'Weighted (pseudo) $C_l$s: data vs model')
        plt.legend(loc = 'lower right')
        plt.xlabel(r'$l$')
        plt.ylabel(r'$l(l+1)C_l/(2\pi)$, arbitrary normalization')
        plt.yscale('log')
        plot.savefig(filename)

        filename = (filename_prefix + '_pcl_ratio' + suffix) if (filename_prefix is not None) else None
        self._plot_smoothed(self.pcl_data/self.pcl_tot, s, label='data/model')
        plt.title(r'(Data pseudo $C_l$) / (Model pseudo $C_l$)')
        plt.legend(loc='lower right')
        plt.xlabel(r'$l$')
        plot.savefig(filename)

        filename = (filename_prefix + '_cl' + suffix) if (filename_prefix is not None) else None
        self._plot_smoothed(self.cl_tot, s, label='model (total)', l2=True)
        self._plot_smoothed(self.cl_beamed_cmb, s, label='model (cmb)', l2=True)
        self._plot_smoothed(self.cl_red_noise, s, label='model (red)', l2=True)
        self._plot_smoothed(self.cl_white_noise, s, label='model (white)', l2=True)
        plt.title(r'Contributions to model $C_l$s (normalized to $\mu$K$^2$)')
        plt.legend(loc = 'lower right')
        plt.xlabel(r'$l$')
        plt.ylabel(r'$l(l+1)C_l/(2\pi)$ ($\mu$K$^2$)')
        plt.yscale('log')
        plot.savefig(filename)      


    def _do_fit(self, cosmo, weight_map):
        """Fits model parameters to self.pcl_data.

        Initializes or updates:

           self.alpha           # current value of alpha
           self.red_ampl        # current value of A in C_l = A (l/lmin)^alpha
           self.white_ampl      # current value of B in C_l = B
           self.cmb_ampl        # multiplier of C_l^{lensed}
           self.red_template    # pseudo-cls corresponding to (l/lmin)^alpha, from one sim
           self.cmb_template    # pseudo-cls corresponding to C_l^{lensed}, from one sim
           self.l_knee          # value of l where white/red noise are equal
           self.uK_arcmin       # reparameterization of self.white_ampl

        Assumes caller has initialized self.alpha. If we're doing iterative fitting, and this
        is not the first call to _do_fit(), then (self.red_template, self.cmb_template) will
        also be used.
        """

        lmin, lmax, lpad = self.lmin, self.lmax, self.lpad

        cl_red = self._cl_red_noise()
        cl_cmb = cosmo.cltt_len[:(lmax+lpad+1)] * self.bl**2   # beamed
        red_template = getattr(self, 'red_template', cl_red)
        cmb_template = getattr(self, 'cmb_template', cl_cmb[:(lmax+1)])
        
        # ninv = inverse variance of C_l, used to determine weighting in fit.
        ninv = np.zeros(lmax+1)

        if self.uniform_lweight:
            ninv[lmin:] = 1.0
        else:
            ninv[lmin:] = 1.0 / (np.arange(lmin,lmax+1) * self.pcl_data[lmin:]**2)

        if False:
            # d = data vector (pcl_data - cmb_template)
            d = self.pcl_data - cmb_template
        
            # M = template matrix (ntemplates by (lmax+1))
            m = [ red_template, white_template ]
            m += [ cmb_template ] if self.fit_cmb_amplitude else []
            m = np.array(m)
            
            a = np.dot(m*ninv, m.T)
            v = np.dot(m*ninv, d)
            coeffs = np.dot(np.linalg.inv(a), v)

            self.red_ampl = coeffs[0]
            self.white_ampl = coeffs[1]
            self.cmb_ampl = (coeffs[2] + 1.0) if self.fit_cmb_amplitude else 1.0

        # Parameter ordering is (red, white, cmb-1, alpha, lred)
        x0 = [0,0]
        x0 += ([0] if self.fit_cmb_amplitude else [])
        x0 += ([self.alpha] if self.fit_alpha else [])
        x0 += ([self.lred] if self.fit_lred else [])
        
        def residual_cl(x):
            alpha = self.alpha
            lred = self.lred
            
            if self.fit_lred:
                lred = x[-1]
                x = x[:-1]

            if self.fit_alpha:
                alpha = x[-1]
                x = x[:-1]

            rcl = self.pcl_data - cmb_template

            # Note: calling _cl_red_noise() at current (alpha,lred), not (self.alpha,self.lred).
            red_cl_ratio = self._cl_red_noise(alpha=alpha,lred=lred) / cl_red
            rcl -= x[0] * red_cl_ratio * red_template
            rcl -= x[1]   # white noise template is C_l = 1

            if self.fit_cmb_amplitude:
                rcl -= x[2] * cmb_template

            return rcl * np.sqrt(ninv)

        lsq = scipy.optimize.least_squares(residual_cl, x0)
        x1 = list(np.copy(lsq['x']))

        self.red_ampl = x1.pop(0)
        self.white_ampl = x1.pop(0)
        self.cmb_ampl = (x1.pop(0)+1) if self.fit_cmb_amplitude else 1.0
        self.alpha = x1.pop(0) if self.fit_alpha else self.alpha
        self.lred = x1.pop(0) if self.fit_lred else self.lred
        self.l_knee = lmin * (self.white_ampl / self.red_ampl)**(1/self.alpha)
        self.uK_arcmin = np.sqrt(self.white_ampl) * (60*180/np.pi)

        # Recompute 'cl_red', in case (alpha, lred) have changed.
        cl_red = self._cl_red_noise(lmax = lmax+lpad)
            
        if not hasattr(self, 'cmb_template'):
            self.cmb_template = self._pcl_from_cl(cl_cmb, weight_map)
            
        if not hasattr(self, 'red_template') or self.fit_alpha:
            self.red_template = self._pcl_from_cl(cl_red, weight_map)

        print('CmbClFitter._do_fit()')
        print(f'    alpha = {self.alpha}')
        print(f'    lred = {self.lred}')
        print(f'    cmb_ampl = {self.cmb_ampl}')
        print(f'    white_ampl = {self.white_ampl}  (uK_arcmin = {self.uK_arcmin})')
        print(f'    red_ampl = {self.red_ampl}    (l_knee = {self.l_knee})')


    def _init_white_noise(self, weight_map, ivar=None):
        """Helper method called by constructor.
        
        self.w2 = sum(weight_map^2 * pixsize) / (4pi)
                = approximate normalization used to compute pseudo-Cls

        self.w2_ivar = sum(weight_map^2 * pixsize^2 / ivar) / (4pi)

        self.ivar_uK_arcmin = sqrt(w2_ivar / w2) * (60*180/pi)
                            = equivalent noise level of ivar map
        """

        wp = weight_map * weight_map.pixsizemap()
        self.w2 = np.vdot(weight_map, wp) / (4*np.pi)
        
        if ivar is None:
            self.w2_ivar = None
            return
        
        invalid = (ivar <= 0)
        
        if np.any(weight_map * invalid):
            raise RuntimeError("CmbClFitter: weight map is nonzero in pixel with ivar <= 0")
        
        wpi = wp / np.where(invalid, 1.0, ivar)   # weight_map * pixsize / ivar
        w2_ivar = np.vdot(wp, wpi) / (4*np.pi)
        self.ivar_uK_arcmin = np.sqrt(w2_ivar/self.w2) * 60*180/np.pi
        print(f'CmbClFitter: equivalent noise level of ivar map = {self.ivar_uK_arcmin} uK-arcmin')
        

    def _cl_red_noise(self, *, alpha=None, lred=None, lmax=None):
        if alpha is None:
            alpha = self.alpha
        if lred is None:
            lred = self.lred
        if lmax is None:
            lmax = self.lmax

        cl = np.maximum(np.arange(lmax+1)/lred, 1.0)**(alpha)
        
        if self.beamed_red_noise:
            cl *= self.bl[:(lmax+1)]**2

        return cl


    def _pcl_from_weighted_map(self, wm):
        alm = pixell_utils.map2alm(wm, self.lmax)
        cl = pixell.curvedsky.alm2cl(alm)
        return cl / self.w2


    def _pcl_from_cl(self, cl_true, weight_map):
        """Simulate map with power spectrum 'cl_true', apply weight map, return (pseudo) Cls.
        
        Some day, I'll implement the pseudo-Cl transfer matrix, and this simulation-based
        approach won't be needed."""

        assert cl_true.shape == ((self.lmax + self.lpad + 1),)
        alm = pixell.curvedsky.rand_alm(cl_true)
        m = pixell_utils.alm2map(alm, weight_map.shape, weight_map.wcs)
        m *= weight_map
        return self._pcl_from_weighted_map(m)

        
    def _plot_smoothed(self, y, s, label=None, l2=False):
        """Helper method for plot()."""
        
        assert y.shape == (self.lmax+1,)
        l = np.arange(self.lmax+1)
        y = (y*l*(l+1)/(2*np.pi)) if l2 else y
        l_smoothed = utils.boxcar_sum(l[self.lmin:], s, normalize=True)
        y_smoothed = utils.boxcar_sum(y[self.lmin:], s, normalize=True)
        plt.plot(l_smoothed, y_smoothed, label=label)
